@@ -2,7 +2,7 @@
 
 set -ue
 
-# Copyright (C) 2017-2024 MariaDB
+# Copyright (C) 2017-2022 MariaDB
 # Copyright (C) 2013 Percona Inc
 #
 # This program is free software; you can redistribute it and/or modify
@@ -19,28 +19,14 @@ set -ue
 # Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston
 # MA  02110-1335  USA.
 
-# This is a reference script for mariadb-backup-based state snapshot transfer.
-
 # Documentation:
 # https://mariadb.com/kb/en/mariabackup-overview/
 # Make sure to read that before proceeding!
 
+OS="$(uname)"
+
 . $(dirname "$0")/wsrep_sst_common
-
-BACKUP_BIN=$(commandex 'mariadb-backup')
-if [ -z "$BACKUP_BIN" ]; then
-    wsrep_log_error 'mariadb-backup binary not found in path'
-    exit 42
-fi
-
-BACKUP_PID=""
-
-INFO_FILE='xtrabackup_galera_info'
-DONOR_INFO_FILE='donor_galera_info'
-IST_FILE='xtrabackup_ist'
-
-MAGIC_FILE="$DATA/$INFO_FILE"
-DONOR_MAGIC_FILE="$DATA/$DONOR_INFO_FILE"
+wsrep_check_datadir
 
 ealgo=""
 eformat=""
@@ -49,6 +35,7 @@ ekeyfile=""
 encrypt=0
 ssyslog=""
 ssystag=""
+BACKUP_PID=""
 tcert=""
 tcap=""
 tpem=""
@@ -77,10 +64,10 @@ tcmd=""
 payload=0
 pvformat="-F '%N => Rate:%r Avg:%a Elapsed:%t %e Bytes: %b %p'"
 pvopts="-f -i 10 -N $WSREP_SST_OPT_ROLE"
+STATDIR=""
 uextra=0
 disver=""
 
-STATDIR=""
 tmpopts=""
 itmpdir=""
 xtmpdir=""
@@ -106,6 +93,22 @@ readonly TOTAL_TAG='total'
 # For backup locks it is 1 sent by joiner
 sst_ver=1
 
+declare -a RC
+
+BACKUP_BIN=$(commandex 'mariadb-backup')
+if [ -z "$BACKUP_BIN" ]; then
+    wsrep_log_error 'mariadb-backup binary not found in path'
+    exit 42
+fi
+
+DATA="$WSREP_SST_OPT_DATA"
+
+INFO_FILE='xtrabackup_galera_info'
+DONOR_INFO_FILE='donor_galera_info'
+IST_FILE='xtrabackup_ist'
+MAGIC_FILE="$DATA/$INFO_FILE"
+DONOR_MAGIC_FILE="$DATA/$DONOR_INFO_FILE"
+
 INNOAPPLYLOG="$DATA/mariabackup.prepare.log"
 INNOMOVELOG="$DATA/mariabackup.move.log"
 INNOBACKUPLOG="$DATA/mariabackup.backup.log"
@@ -439,30 +442,17 @@ get_transfer()
 get_footprint()
 {
     cd "$DATA_DIR"
-    local payload_data
-    if [ "$OS" = 'Linux' ]; then
-        payload_data=$(find $findopt . \
-            -regex '.*undo[0-9]+$\|.*\.ibd$\|.*\.MYI$\|.*\.MYD$\|.*ibdata1$' \
-            -type f -print0 | du --files0-from=- --bytes -c -s | \
-            awk 'END { print $1 }')
-    else
-        payload_data=$(find $findopt . \
-            -regex '.*undo[0-9]+$|.*\.ibd$|.*\.MYI$\.*\.MYD$|.*ibdata1$' \
-            -type f -print0 | xargs -0 stat -f '%z' | \
-            awk '{ sum += $1 } END { print sum }')
-    fi
+    local payload_data=$(find $findopt . \
+        -regex '.*undo[0-9]+$\|.*\.ibd$\|.*\.MYI$\|.*\.MYD$\|.*ibdata1$' \
+        -type f -print0 | du --files0-from=- --block-size=1 -c -s | \
+        awk 'END { print $1 }')
     local payload_undo=0
     if [ -n "$ib_undo_dir" -a "$ib_undo_dir" != '.' -a \
          "$ib_undo_dir" != "$DATA_DIR" -a -d "$ib_undo_dir" ]
     then
         cd "$ib_undo_dir"
-        if [ "$OS" = 'Linux' ]; then
-            payload_undo=$(find . -regex '.*undo[0-9]+$' -type f -print0 | \
-                du --files0-from=- --bytes -c -s | awk 'END { print $1 }')
-        else
-            payload_undo=$(find . -regex '.*undo[0-9]+$' -type f -print0 | \
-                xargs -0 stat -f '%z' | awk '{ sum += $1 } END { print sum }')
-        fi
+        payload_undo=$(find . -regex '.*undo[0-9]+$' -type f -print0 | \
+            du --files0-from=- --block-size=1 -c -s | awk 'END { print $1 }')
     fi
     cd "$OLD_PWD"
 
@@ -540,8 +530,7 @@ adjust_progress()
     fi
 }
 
-bkgroups='sst|xtrabackup|mariabackup'
-encgroups="--mysqld|$bkgroups"
+encgroups='--mysqld|sst|xtrabackup'
 
 read_cnf()
 {
@@ -597,34 +586,26 @@ read_cnf()
         ssl_dhparams=$(parse_cnf "$encgroups" 'ssl-dhparams')
     fi
 
-    sockopt=$(parse_cnf sst sockopt)
-    progress=$(parse_cnf sst progress)
+    sockopt=$(parse_cnf sst sockopt "")
+    progress=$(parse_cnf sst progress "")
     ttime=$(parse_cnf sst time 0)
     cpat='.*\.pem$\|.*galera\.cache$\|.*sst_in_progress$\|.*\.sst$\|.*gvwstate\.dat$\|.*grastate\.dat$\|.*\.err$\|.*\.log$\|.*RPM_UPGRADE_MARKER$\|.*RPM_UPGRADE_HISTORY$'
     [ "$OS" = 'FreeBSD' ] && cpat=$(echo "$cpat" | sed 's/\\|/|/g')
     cpat=$(parse_cnf sst cpat "$cpat")
-    scomp=$(parse_cnf sst compressor)
-    sdecomp=$(parse_cnf sst decompressor)
+    scomp=$(parse_cnf sst compressor "")
+    sdecomp=$(parse_cnf sst decompressor "")
 
-    rlimit=$(parse_cnf sst rlimit)
+    rlimit=$(parse_cnf sst rlimit "")
     uextra=$(parse_cnf sst use-extra 0)
-    speciald=$(parse_cnf sst 'sst-special-dirs' 1)
-    iopts=$(parse_cnf "$bkgroups" 'inno-backup-opts')
-    iapts=$(parse_cnf "$bkgroups" 'inno-apply-opts')
-    impts=$(parse_cnf "$bkgroups" 'inno-move-opts')
-    use_memory=$(parse_cnf "$bkgroups" 'use-memory')
-    if [ -z "$use_memory" ]; then
-        if [ -n "$INNODB_BUFFER_POOL_SIZE" ]; then
-            use_memory="$INNODB_BUFFER_POOL_SIZE"
-        else
-            use_memory=$(parse_cnf '--mysqld' 'innodb-buffer-pool-size')
-        fi
-    fi
-    stimeout=$(parse_cnf sst 'sst-initial-timeout' 300)
-    ssyslog=$(parse_cnf sst 'sst-syslog' 0)
-    ssystag=$(parse_cnf mysqld_safe 'syslog-tag' "${SST_SYSLOG_TAG:-}")
+    speciald=$(parse_cnf sst sst-special-dirs 1)
+    iopts=$(parse_cnf sst inno-backup-opts "")
+    iapts=$(parse_cnf sst inno-apply-opts "")
+    impts=$(parse_cnf sst inno-move-opts "")
+    stimeout=$(parse_cnf sst sst-initial-timeout 300)
+    ssyslog=$(parse_cnf sst sst-syslog 0)
+    ssystag=$(parse_cnf mysqld_safe syslog-tag "${SST_SYSLOG_TAG:-}")
     ssystag="$ssystag-"
-    sstlogarchive=$(parse_cnf sst 'sst-log-archive' 1)
+    sstlogarchive=$(parse_cnf sst sst-log-archive 1)
     sstlogarchivedir=""
     if [ $sstlogarchive -ne 0 ]; then
         sstlogarchivedir=$(parse_cnf sst sst-log-archive-dir \
@@ -698,25 +679,24 @@ cleanup_at_exit()
 
     [ "$(pwd)" != "$OLD_PWD" ] && cd "$OLD_PWD"
 
-    if [ "$WSREP_SST_OPT_ROLE" = 'donor' -o $estatus -ne 0 ]; then
-        if [ $estatus -ne 0 ]; then
-            wsrep_log_error "Removing $MAGIC_FILE file due to signal"
-        fi
+    if [ $estatus -ne 0 ]; then
+        wsrep_log_error "Removing $MAGIC_FILE file due to signal"
         [ -f "$MAGIC_FILE" ] && rm -f "$MAGIC_FILE" || :
         [ -f "$DONOR_MAGIC_FILE" ] && rm -f "$DONOR_MAGIC_FILE" || :
-        [ -f "$DATA/$IST_FILE" ] && rm -f "$DATA/$IST_FILE" || :
     fi
 
     if [ "$WSREP_SST_OPT_ROLE" = 'joiner' ]; then
-        if [ -n "$BACKUP_PID" ]; then
-            if ps -p $BACKUP_PID >/dev/null 2>&1; then
-                wsrep_log_error \
-                    "SST streaming process is still running. Killing..."
-                cleanup_pid $BACKUP_PID
-            fi
-        fi
         wsrep_log_info "Removing the sst_in_progress file"
         wsrep_cleanup_progress_file
+    else
+        if [ -n "$BACKUP_PID" ]; then
+            if check_pid "$BACKUP_PID" 1; then
+                wsrep_log_error \
+                    "mariadb-backup process is still running. Killing..."
+                cleanup_pid $CHECK_PID "$BACKUP_PID"
+            fi
+        fi
+        [ -f "$DATA/$IST_FILE" ] && rm -f "$DATA/$IST_FILE" || :
     fi
 
     if [ -n "$progress" -a -p "$progress" ]; then
@@ -734,7 +714,7 @@ cleanup_at_exit()
     fi
 
     # Final cleanup
-    local pgid=$(ps -o 'pgid=' $$ 2>/dev/null | grep -o -E '[0-9]+' || :)
+    pgid=$(ps -o pgid= $$ 2>/dev/null | grep -o -E '[0-9]+' || :)
 
     # This means no setsid done in mysqld.
     # We don't want to kill mysqld here otherwise.
@@ -748,7 +728,7 @@ cleanup_at_exit()
         fi
     fi
 
-    if [ -n "$SST_PID" ]; then
+    if [ -n "${SST_PID:-}" ]; then
         [ -f "$SST_PID" ] && rm -f "$SST_PID" || :
     fi
 
@@ -772,7 +752,7 @@ setup_ports()
 wait_for_listen()
 {
     for i in {1..150}; do
-        if check_port "" "$SST_PORT" 'socat|nc|netcat'; then
+        if check_port "" "$SST_PORT" 'socat|nc'; then
             break
         fi
         sleep 0.2
@@ -871,7 +851,7 @@ recv_joiner()
             wsrep_log_error "receiving process ended without creating" \
                             "magic file ($MAGIC_FILE)"
             wsrep_log_info "Contents of datadir:"
-            wsrep_log_info "$(ls -l "$dir"/*)"
+            wsrep_log_info $(ls -l "$dir/"*)
             exit 32
         fi
 
@@ -925,20 +905,23 @@ monitor_process()
     local sst_stream_pid=$1
 
     while :; do
-        if ! ps -p $WSREP_SST_OPT_PARENT >/dev/null 2>&1; then
+        if ! ps -p "$WSREP_SST_OPT_PARENT" >/dev/null 2>&1; then
             wsrep_log_error \
                 "Parent mysqld process (PID: $WSREP_SST_OPT_PARENT)" \
                 "terminated unexpectedly."
-            kill -- -$WSREP_SST_OPT_PARENT
+            kill -- -"$WSREP_SST_OPT_PARENT"
             exit 32
         fi
-        if ! ps -p $sst_stream_pid >/dev/null 2>&1; then
+        if ! ps -p "$sst_stream_pid" >/dev/null 2>&1; then
             break
         fi
         sleep 0.1
     done
 }
 
+[ -f "$MAGIC_FILE" ] && rm -f "$MAGIC_FILE"
+[ -f "$DONOR_MAGIC_FILE" ] && rm -rf "$DONOR_MAGIC_FILE"
+
 read_cnf
 setup_ports
 
@@ -954,17 +937,15 @@ if "$BACKUP_BIN" --help 2>/dev/null | grep -qw -F -- '--version-check'; then
     disver=' --no-version-check'
 fi
 
-get_stream
-get_transfer
+OLD_PWD="$(pwd)"
 
-findopt='-L'
-[ "$OS" = 'FreeBSD' ] && findopt="$findopt -E"
+if [ -n "$DATA" -a "$DATA" != '.' ]; then
+    [ ! -d "$DATA" ] && mkdir -p "$DATA"
+    cd "$DATA"
+fi
+DATA_DIR="$(pwd)"
 
-wait_previous_sst
-
-[ -f "$MAGIC_FILE" ] && rm -f "$MAGIC_FILE"
-[ -f "$DONOR_MAGIC_FILE" ] && rm -f "$DONOR_MAGIC_FILE"
-[ -f "$DATA/$IST_FILE" ] && rm -f "$DATA/$IST_FILE"
+cd "$OLD_PWD"
 
 if [ $ssyslog -eq 1 ]; then
     if [ -n "$(commandex logger)" ]; then
@@ -974,17 +955,17 @@ if [ $ssyslog -eq 1 ]; then
 
         wsrep_log_error()
         {
-            logger -p daemon.err -t ${ssystag}wsrep-sst-$WSREP_SST_OPT_ROLE -- "$@"
+            logger -p daemon.err -t ${ssystag}wsrep-sst-$WSREP_SST_OPT_ROLE "$@"
         }
 
         wsrep_log_warning()
         {
-            logger -p daemon.warning -t ${ssystag}wsrep-sst-$WSREP_SST_OPT_ROLE -- "$@"
+            logger -p daemon.warning -t ${ssystag}wsrep-sst-$WSREP_SST_OPT_ROLE "$@"
         }
 
         wsrep_log_info()
         {
-            logger -p daemon.info -t ${ssystag}wsrep-sst-$WSREP_SST_OPT_ROLE -- "$@"
+            logger -p daemon.info -t ${ssystag}wsrep-sst-$WSREP_SST_OPT_ROLE "$@"
         }
     else
         wsrep_log_error "logger not in path: $PATH. Ignoring"
@@ -1062,9 +1043,6 @@ setup_commands()
     if [ -n "$INNODB_FORCE_RECOVERY" ]; then
         recovery=" --innodb-force-recovery=$INNODB_FORCE_RECOVERY"
     fi
-    if [ -n "$use_memory" ]; then
-        INNOEXTRA="$INNOEXTRA --use-memory=$use_memory"
-    fi
     INNOAPPLY="$BACKUP_BIN --prepare$disver$recovery${iapts:+ }$iapts$INNOEXTRA --target-dir='$DATA' --datadir='$DATA'$mysqld_args $INNOAPPLY"
     INNOMOVE="$BACKUP_BIN$WSREP_SST_OPT_CONF --move-back$disver${impts:+ }$impts$INNOEXTRA --galera-info --force-non-empty-directories --target-dir='$DATA' --datadir='${TDATA:-$DATA}' $INNOMOVE"
     INNOBACKUP="$BACKUP_BIN$WSREP_SST_OPT_CONF --backup$disver${iopts:+ }$iopts$tmpopts$INNOEXTRA --galera-info --stream=$sfmt --target-dir='$itmpdir' --datadir='$DATA'$mysqld_args $INNOBACKUP"
@@ -1087,6 +1065,12 @@ send_magic()
     fi
 }
 
+get_stream
+get_transfer
+
+findopt='-L'
+[ "$OS" = 'FreeBSD' ] && findopt="$findopt -E"
+
 if [ "$WSREP_SST_OPT_ROLE" = 'donor' ]; then
 
     trap cleanup_at_exit EXIT
@@ -1102,27 +1086,27 @@ if [ "$WSREP_SST_OPT_ROLE" = 'donor' ]; then
         tmpdir=$(parse_cnf "$encgroups" 'tmpdir')
         if [ -z "$tmpdir" ]; then
             xtmpdir="$(mktemp -d)"
-            itmpdir="$(mktemp -d)"
         elif [ "$OS" = 'Linux' ]; then
-            xtmpdir=$(mktemp -d "--tmpdir=$tmpdir")
-            itmpdir=$(mktemp -d "--tmpdir=$tmpdir")
+            xtmpdir=$(mktemp '-d' "--tmpdir=$tmpdir")
         else
-            xtmpdir=$(TMPDIR="$tmpdir"; mktemp -d)
-            itmpdir=$(TMPDIR="$tmpdir"; mktemp -d)
+            xtmpdir=$(TMPDIR="$tmpdir"; mktemp '-d')
         fi
 
         wsrep_log_info "Using '$xtmpdir' as mariadb-backup temporary directory"
         tmpopts=" --tmpdir='$xtmpdir'"
 
-        wsrep_log_info "Using '$itmpdir' as mariadb-backup working directory"
+        itmpdir="$(mktemp -d)"
+        wsrep_log_info "Using '$itmpdir' as mariadb-abackup working directory"
 
+        usrst=0
         if [ -n "$WSREP_SST_OPT_USER" ]; then
            INNOEXTRA="$INNOEXTRA --user='$WSREP_SST_OPT_USER'"
+           usrst=1
         fi
 
         if [ -n "$WSREP_SST_OPT_PSWD" ]; then
             export MYSQL_PWD="$WSREP_SST_OPT_PSWD"
-        elif [ -n "$WSREP_SST_OPT_USER" ]; then
+        elif [ $usrst -eq 1 ]; then
             # Empty password, used for testing, debugging etc.
             unset MYSQL_PWD
         fi
@@ -1214,14 +1198,17 @@ if [ "$WSREP_SST_OPT_ROLE" = 'donor' ]; then
             exit 22
         fi
 
+        # mariadb-backup implicitly writes PID to fixed location in $xtmpdir
+        BACKUP_PID="$xtmpdir/xtrabackup_pid"
+
     else # BYPASS FOR IST
 
         wsrep_log_info "Bypassing the SST for IST"
-        echo 'continue' # now server can resume updating data
+        echo "continue" # now server can resume updating data
 
         send_magic
 
-        echo '1' > "$DATA/$IST_FILE"
+        echo "1" > "$DATA/$IST_FILE"
 
         if [ -n "$scomp" ]; then
             tcmd="$scomp | $tcmd"
@@ -1243,26 +1230,132 @@ if [ "$WSREP_SST_OPT_ROLE" = 'donor' ]; then
 
 else # joiner
 
-    create_dirs 1
-
     [ -e "$SST_PROGRESS_FILE" ] && \
         wsrep_log_info "Stale sst_in_progress file: $SST_PROGRESS_FILE"
     [ -n "$SST_PROGRESS_FILE" ] && touch "$SST_PROGRESS_FILE"
 
+    # if no command line argument and INNODB_DATA_HOME_DIR environment
+    # variable is not set, try to get it from the my.cnf:
+    if [ -z "$INNODB_DATA_HOME_DIR" ]; then
+        INNODB_DATA_HOME_DIR=$(parse_cnf '--mysqld' 'innodb-data-home-dir')
+        INNODB_DATA_HOME_DIR=$(trim_dir "$INNODB_DATA_HOME_DIR")
+    fi
+
+    if [ -n "$INNODB_DATA_HOME_DIR" -a "$INNODB_DATA_HOME_DIR" != '.' -a \
+         "$INNODB_DATA_HOME_DIR" != "$DATA_DIR" ]
+    then
+        # handle both relative and absolute paths:
+        cd "$DATA"
+        [ ! -d "$INNODB_DATA_HOME_DIR" ] && mkdir -p "$INNODB_DATA_HOME_DIR"
+        cd "$INNODB_DATA_HOME_DIR"
+        ib_home_dir="$(pwd)"
+        cd "$OLD_PWD"
+        [ "$ib_home_dir" = "$DATA_DIR" ] && ib_home_dir=""
+    fi
+
+    # if no command line argument and INNODB_LOG_GROUP_HOME is not set,
+    # then try to get it from the my.cnf:
+    if [ -z "$INNODB_LOG_GROUP_HOME" ]; then
+        INNODB_LOG_GROUP_HOME=$(parse_cnf '--mysqld' 'innodb-log-group-home-dir')
+        INNODB_LOG_GROUP_HOME=$(trim_dir "$INNODB_LOG_GROUP_HOME")
+    fi
+
+    if [ -n "$INNODB_LOG_GROUP_HOME" -a "$INNODB_LOG_GROUP_HOME" != '.' -a \
+         "$INNODB_LOG_GROUP_HOME" != "$DATA_DIR" ]
+    then
+        # handle both relative and absolute paths:
+        cd "$DATA"
+        [ ! -d "$INNODB_LOG_GROUP_HOME" ] && mkdir -p "$INNODB_LOG_GROUP_HOME"
+        cd "$INNODB_LOG_GROUP_HOME"
+        ib_log_dir="$(pwd)"
+        cd "$OLD_PWD"
+        [ "$ib_log_dir" = "$DATA_DIR" ] && ib_log_dir=""
+    fi
+
+    # if no command line argument and INNODB_UNDO_DIR is not set,
+    # then try to get it from the my.cnf:
+    if [ -z "$INNODB_UNDO_DIR" ]; then
+        INNODB_UNDO_DIR=$(parse_cnf '--mysqld' 'innodb-undo-directory')
+        INNODB_UNDO_DIR=$(trim_dir "$INNODB_UNDO_DIR")
+    fi
+
+    if [ -n "$INNODB_UNDO_DIR" -a "$INNODB_UNDO_DIR" != '.' -a \
+         "$INNODB_UNDO_DIR" != "$DATA_DIR" ]
+    then
+        # handle both relative and absolute paths:
+        cd "$DATA"
+        [ ! -d "$INNODB_UNDO_DIR" ] && mkdir -p "$INNODB_UNDO_DIR"
+        cd "$INNODB_UNDO_DIR"
+        ib_undo_dir="$(pwd)"
+        cd "$OLD_PWD"
+        [ "$ib_undo_dir" = "$DATA_DIR" ] && ib_undo_dir=""
+    fi
+
+    # if no command line argument then try to get it from the my.cnf:
+    if [ -z "$ARIA_LOG_DIR" ]; then
+        ARIA_LOG_DIR=$(parse_cnf '--mysqld' 'aria-log-dir-path')
+        ARIA_LOG_DIR=$(trim_dir "$ARIA_LOG_DIR")
+    fi
+
+    if [ -n "$ARIA_LOG_DIR" -a "$ARIA_LOG_DIR" != '.' -a \
+         "$ARIA_LOG_DIR" != "$DATA_DIR" ]
+    then
+        # handle both relative and absolute paths:
+        cd "$DATA"
+        [ ! -d "$ARIA_LOG_DIR" ] && mkdir -p "$ARIA_LOG_DIR"
+        cd "$ARIA_LOG_DIR"
+        ar_log_dir="$(pwd)"
+        cd "$OLD_PWD"
+        [ "$ar_log_dir" = "$DATA_DIR" ] && ar_log_dir=""
+    fi
+
     if [ -n "$backup_threads" ]; then
         impts="--parallel=$backup_threads${impts:+ }$impts"
     fi
 
+    SST_PID="$DATA/wsrep_sst.pid"
+
+    # give some time for previous SST to complete:
+    check_round=0
+    while check_pid "$SST_PID" 0; do
+        wsrep_log_info "previous SST is not completed, waiting for it to exit"
+        check_round=$(( check_round + 1 ))
+        if [ $check_round -eq 10 ]; then
+            wsrep_log_error "previous SST script still running."
+            exit 114 # EALREADY
+        fi
+        sleep 1
+    done
+
+    trap simple_cleanup EXIT
+    echo $$ > "$SST_PID"
+
     stagemsg='Joiner-Recv'
 
     MODULE="${WSREP_SST_OPT_MODULE:-xtrabackup_sst}"
+
+    [ -f "$DATA/$IST_FILE" ] && rm -f "$DATA/$IST_FILE"
+
+    # May need xtrabackup_checkpoints later on
+    [ -f "$DATA/xtrabackup_binary" ]      && rm -f "$DATA/xtrabackup_binary"
+    [ -f "$DATA/xtrabackup_galera_info" ] && rm -f "$DATA/xtrabackup_galera_info"
+
     ADDR="$WSREP_SST_OPT_HOST"
 
     if [ "${tmode#VERIFY}" != "$tmode" ]; then
         # backward-incompatible behavior:
         CN=""
         if [ -n "$tpem" ]; then
-            CN=$(openssl_getCN "$tpem")
+            # find out my Common Name
+            get_openssl
+            if [ -z "$OPENSSL_BINARY" ]; then
+                wsrep_log_error \
+                    'openssl not found but it is required for authentication'
+                exit 42
+            fi
+            CN=$("$OPENSSL_BINARY" x509 -noout -subject -in "$tpem" | \
+                 tr ',' '\n' | grep -F 'CN =' | cut -d '=' -f2 | sed s/^\ // | \
+                 sed s/\ %//)
         fi
         MY_SECRET="$(wsrep_gen_secret)"
         # Add authentication data to address
@@ -1286,7 +1379,6 @@ else # joiner
 
     STATDIR="$(mktemp -d)"
     MAGIC_FILE="$STATDIR/$INFO_FILE"
-    DONOR_MAGIC_FILE="$STATDIR/$DONOR_INFO_FILE"
 
     recv_joiner "$STATDIR" "$stagemsg-gtid" $stimeout 1 1
 
@@ -1317,7 +1409,7 @@ else # joiner
         fi
         mkdir -p "$DATA/.sst"
         (recv_joiner "$DATA/.sst" "$stagemsg-SST" 0 0 0) &
-        BACKUP_PID=$!
+        jpid=$!
         wsrep_log_info "Proceeding with SST"
 
         get_binlog
@@ -1360,22 +1452,12 @@ else # joiner
                 "$DATA" -mindepth 1 -prune -regex "$cpat" \
                 -o -exec rm -rf {} >&2 \+
 
-        # Deleting files from previous SST and legacy files from old versions:
-        [ -f "$DATA/xtrabackup_binary" ]      && rm -f "$DATA/xtrabackup_binary"
-        [ -f "$DATA/xtrabackup_pid" ]         && rm -f "$DATA/xtrabackup_pid"
-        [ -f "$DATA/xtrabackup_checkpoints" ] && rm -f "$DATA/xtrabackup_checkpoints"
-        [ -f "$DATA/xtrabackup_info" ]        && rm -f "$DATA/xtrabackup_info"
-        [ -f "$DATA/xtrabackup_slave_info" ]  && rm -f "$DATA/xtrabackup_slave_info"
-        [ -f "$DATA/xtrabackup_binlog_info" ] && rm -f "$DATA/xtrabackup_binlog_info"
-        [ -f "$DATA/xtrabackup_binlog_pos_innodb" ] && rm -f "$DATA/xtrabackup_binlog_pos_innodb"
-
         TDATA="$DATA"
         DATA="$DATA/.sst"
         MAGIC_FILE="$DATA/$INFO_FILE"
 
         wsrep_log_info "Waiting for SST streaming to complete!"
-        monitor_process $BACKUP_PID
-        BACKUP_PID=""
+        monitor_process $jpid
 
         if [ ! -s "$DATA/xtrabackup_checkpoints" ]; then
             wsrep_log_error "xtrabackup_checkpoints missing," \
@@ -1386,7 +1468,7 @@ else # joiner
         # Compact backups are not supported by mariadb-backup
         if grep -qw -F 'compact = 1' "$DATA/xtrabackup_checkpoints"; then
             wsrep_log_info "Index compaction detected"
-            wsrep_log_error "Compact backups are not supported by mariadb-backup"
+            wsrel_log_error "Compact backups are not supported by mariadb-backup"
             exit 2
         fi
 
@@ -1483,7 +1565,6 @@ else # joiner
         fi
 
         MAGIC_FILE="$TDATA/$INFO_FILE"
-        DONOR_MAGIC_FILE="$TDATA/$DONOR_INFO_FILE"
 
         wsrep_log_info "Moving the backup to $TDATA"
         timeit 'mariadb-backup move stage' "$INNOMOVE"
@@ -1508,13 +1589,10 @@ else # joiner
     fi
 
     if [ ! -r "$MAGIC_FILE" ]; then
-        wsrep_log_error "Internal error: SST magic file '$MAGIC_FILE'" \
-                        "not found or not readable"
+        wsrep_log_error "SST magic file '$MAGIC_FILE' not found/readable"
         exit 2
     fi
 
-    simulate_long_sst
-
     # use donor magic file, if present
     # if IST was used, donor magic file was not created
     # Remove special tags from the magic file, and from the output:
