@@ -104,7 +104,7 @@ typedef struct st_federatedx_share {
   /*
     the primary select query to be used in rnd_init
   */
-  char *select_query;
+  LEX_CSTRING select_query;
   /*
     remote host info, parse_url supplies
   */
@@ -222,7 +222,6 @@ public:
   virtual int seek_position(FEDERATEDX_IO_RESULT **io_result,
                             const void *ref)=0;
   virtual void set_thd(void *thd) { }
-
 };
 
 
@@ -258,7 +257,6 @@ public:
   void stmt_autocommit();
 };
 
-
 /*
   Class definition for the storage engine
 */
@@ -292,12 +290,10 @@ private:
                                       FEDERATEDX_IO_RESULT *result);
   bool create_where_from_key(String *to, KEY *key_info,
                              const key_range *start_key,
-                             const key_range *end_key,
-                             bool records_in_range, bool eq_range);
+                             const key_range *end_key, bool eq_range);
   int stash_remote_error();
 
-  federatedx_txn *get_txn(THD *thd, bool no_create= FALSE);
-
+  static federatedx_txn *get_txn(THD *thd, bool no_create= FALSE);
   static int disconnect(handlerton *hton, MYSQL_THD thd);
   static int savepoint_set(handlerton *hton, MYSQL_THD thd, void *sv);
   static int savepoint_rollback(handlerton *hton, MYSQL_THD thd, void *sv);
@@ -319,7 +315,7 @@ private:
   int real_connect(FEDERATEDX_SHARE *my_share, uint create_flag);
 public:
   ha_federatedx(handlerton *hton, TABLE_SHARE *table_arg);
-  ~ha_federatedx() {}
+  ~ha_federatedx() = default;
   /*
     The name of the index type that will be used for display
     don't implement this method unless you really have indexes
@@ -368,29 +364,31 @@ public:
     Talk to Kostja about this - how to get the
     number of rows * ...
     disk scan time on other side (block size, size of the row) + network time ...
-    The reason for "records * 1000" is that such a large number forces
-    this to use indexes "
+    The reason for "1000" is that such a large number forces this to use indexes "
   */
-  double scan_time()
+  IO_AND_CPU_COST scan_time()
   {
     DBUG_PRINT("info", ("records %lu", (ulong) stats.records));
-    return (double)(stats.records*1000);
+    return
+    {
+      0,
+        (double) (stats.mean_rec_length * stats.records)/8192 * DISK_READ_COST+
+        1000,
+    };
   }
-  /*
-    The next method will never be called if you do not implement indexes.
-  */
-  double read_time(uint index, uint ranges, ha_rows rows)
+  IO_AND_CPU_COST keyread_time(uint index, ulong ranges, ha_rows rows,
+                               ulonglong blocks)
   {
-    /*
-      Per Brian, this number is bugus, but this method must be implemented,
-      and at a later date, he intends to document this issue for handler code
-    */
-    return (double) rows /  20.0+1;
+    return {0, (double) (ranges + rows) * DISK_READ_COST };
+  }
+  IO_AND_CPU_COST rnd_pos_time(ha_rows rows)
+  {
+    return {0, (double) rows * DISK_READ_COST };
   }
 
   const key_map *keys_to_use_for_scanning() { return &key_map_full; }
   /*
-    Everything below are methods that we implment in ha_federatedx.cc.
+    Everything below are methods that we implement in ha_federatedx.cc.
 
     Most of these methods are not obligatory, skip them and
     MySQL will treat them as not implemented
