@@ -24,31 +24,58 @@
 #include "sql_priv.h"
 #include "sql_plugin.h"
 #include "ha_heap.h"
-#include "sql_base.h"                    // enum_tdc_remove_table_type
+#include "sql_base.h"
 
 static handler *heap_create_handler(handlerton *, TABLE_SHARE *, MEM_ROOT *);
 static int heap_prepare_hp_create_info(TABLE *, bool, HP_CREATE_INFO *);
 
 
-int heap_panic(handlerton *hton, ha_panic_function flag)
+static int heap_panic(handlerton *hton, ha_panic_function flag)
 {
   return hp_panic(flag);
 }
 
 
+static int heap_drop_table(handlerton *hton, const char *path)
+{
+  int error= heap_delete_table(path);
+  return error == ENOENT ? -1 : error;
+}
+
+/* See optimizer_costs.txt for how the following values where calculated */
+#define HEAP_ROW_NEXT_FIND_COST  8.0166e-06           // For table scan
+#define BTREE_KEY_NEXT_FIND_COST 0.00007739           // For binary tree scan
+#define HEAP_LOOKUP_COST         0.00016097           // Heap lookup cost
+
+static void heap_update_optimizer_costs(OPTIMIZER_COSTS *costs)
+{
+  /*
+    A lot of values are 0 as heap supports all needed xxx_time() functions
+  */
+  costs->disk_read_cost=0;          // All data in memory
+  costs->disk_read_ratio= 0.0;      // All data in memory
+  costs->key_next_find_cost= 0;
+  costs->key_copy_cost= 0;          // Set in keyread_time()
+  costs->row_copy_cost= 2.334e-06;  // This is small as its just a memcpy
+  costs->row_lookup_cost= 0;        // Direct pointer
+  costs->row_next_find_cost= 0;
+  costs->key_lookup_cost= 0;
+  costs->key_next_find_cost= 0;
+  costs->index_block_copy_cost= 0;
+}
+
 int heap_init(void *p)
 {
   handlerton *heap_hton;
 
-#ifdef HAVE_PSI_INTERFACE
   init_heap_psi_keys();
-#endif
 
   heap_hton= (handlerton *)p;
-  heap_hton->state=      SHOW_OPTION_YES;
   heap_hton->db_type=    DB_TYPE_HEAP;
   heap_hton->create=     heap_create_handler;
   heap_hton->panic=      heap_panic;
+  heap_hton->drop_table= heap_drop_table;
+  heap_hton->update_optimizer_costs= heap_update_optimizer_costs;
   heap_hton->flags=      HTON_CAN_RECREATE;
 
   return 0;
@@ -69,7 +96,8 @@ static handler *heap_create_handler(handlerton *hton,
 ha_heap::ha_heap(handlerton *hton, TABLE_SHARE *table_arg)
   :handler(hton, table_arg), file(0), records_changed(0), key_stat_version(0), 
   internal_table(0)
-{}
+{
+}
 
 /*
   Hash index statistics is updated (copied from HP_KEYDEF::hash_buckets to 
@@ -149,7 +177,7 @@ int ha_heap::close(void)
 
 handler *ha_heap::clone(const char *name, MEM_ROOT *mem_root)
 {
-  handler *new_handler= get_new_handler(table->s, mem_root, table->s->db_type());
+  handler *new_handler= get_new_handler(table->s, mem_root, ht);
   if (new_handler && !new_handler->ha_open(table, file->s->name, table->db_stat,
                                            HA_OPEN_IGNORE_IF_LOCKED))
     return new_handler;
@@ -224,6 +252,41 @@ void ha_heap::update_key_stats()
 }
 
 
+IO_AND_CPU_COST ha_heap::keyread_time(uint index, ulong ranges, ha_rows rows,
+                                      ulonglong blocks)
+{
+  KEY *key=table->key_info+index;
+  if (key->algorithm == HA_KEY_ALG_BTREE)
+  {
+    double lookup_cost;
+    lookup_cost= ranges * costs->key_cmp_cost * log2(stats.records+1);
+    return {0, ranges * lookup_cost + (rows-ranges) * BTREE_KEY_NEXT_FIND_COST };
+  }
+  else
+  {
+    return {0, (ranges * HEAP_LOOKUP_COST +
+                (rows-ranges) * BTREE_KEY_NEXT_FIND_COST) };
+  }
+}
+
+
+IO_AND_CPU_COST ha_heap::scan_time()
+{
+  return {0, (double) (stats.records+stats.deleted) * HEAP_ROW_NEXT_FIND_COST };
+}
+
+
+IO_AND_CPU_COST ha_heap::rnd_pos_time(ha_rows rows)
+{
+  /*
+    The row pointer is a direct pointer to the block. Thus almost instant
+    in practice.
+    Note that ha_rnd_pos_time() will add ROW_COPY_COST to this result
+  */
+  return { 0, 0 };
+}
+
+
 int ha_heap::write_row(const uchar * buf)
 {
   int res;
@@ -555,8 +618,7 @@ THR_LOCK_DATA **ha_heap::store_lock(THD *thd,
 
 int ha_heap::delete_table(const char *name)
 {
-  int error= heap_delete_table(name);
-  return error == ENOENT ? 0 : error;
+  return heap_drop_table(0, name);
 }
 
 
@@ -573,8 +635,8 @@ int ha_heap::rename_table(const char * from, const char * to)
 }
 
 
-ha_rows ha_heap::records_in_range(uint inx, key_range *min_key,
-                                  key_range *max_key)
+ha_rows ha_heap::records_in_range(uint inx, const key_range *min_key,
+                                  const key_range *max_key, page_range *pages)
 {
   KEY *key=table->key_info+inx;
   if (key->algorithm == HA_KEY_ALG_BTREE)
@@ -612,7 +674,8 @@ static int heap_prepare_hp_create_info(TABLE *table_arg, bool internal_table,
   for (key= parts= 0; key < keys; key++)
     parts+= table_arg->key_info[key].user_defined_key_parts;
 
-  if (!my_multi_malloc(MYF(MY_WME | MY_THREAD_SPECIFIC),
+  if (!my_multi_malloc(hp_key_memory_HP_KEYDEF,
+                       MYF(MY_WME | MY_THREAD_SPECIFIC),
                        &keydef, keys * sizeof(HP_KEYDEF),
                        &seg, parts * sizeof(HA_KEYSEG),
                        NULL))
@@ -827,23 +890,6 @@ int ha_heap::find_unique_row(uchar *record, uint unique_idx)
 struct st_mysql_storage_engine heap_storage_engine=
 { MYSQL_HANDLERTON_INTERFACE_VERSION };
 
-mysql_declare_plugin(heap)
-{
-  MYSQL_STORAGE_ENGINE_PLUGIN,
-  &heap_storage_engine,
-  "MEMORY",
-  "MySQL AB",
-  "Hash based, stored in memory, useful for temporary tables",
-  PLUGIN_LICENSE_GPL,
-  heap_init,
-  NULL,
-  0x0100, /* 1.0 */
-  NULL,                       /* status variables                */
-  NULL,                       /* system variables                */
-  NULL,                       /* config options                  */
-  0,                          /* flags                           */
-}
-mysql_declare_plugin_end;
 maria_declare_plugin(heap)
 {
   MYSQL_STORAGE_ENGINE_PLUGIN,
