@@ -33,7 +33,6 @@ Created 2011/12/19
 #include "trx0sys.h"
 #include "fil0crypt.h"
 #include "fil0pagecompress.h"
-#include "log.h"
 
 using st_::span;
 
@@ -54,6 +53,7 @@ void buf_dblwr_t::init()
     active_slot= &slots[0];
     mysql_mutex_init(buf_dblwr_mutex_key, &mutex, nullptr);
     pthread_cond_init(&cond, nullptr);
+    block_size= FSP_EXTENT_SIZE;
   }
 }
 
@@ -68,7 +68,7 @@ inline void buf_dblwr_t::init(const byte *header)
   block1= page_id_t(0, mach_read_from_4(header + TRX_SYS_DOUBLEWRITE_BLOCK1));
   block2= page_id_t(0, mach_read_from_4(header + TRX_SYS_DOUBLEWRITE_BLOCK2));
 
-  const uint32_t buf_size= 2 * block_size();
+  const uint32_t buf_size= 2 * block_size;
   for (int i= 0; i < 2; i++)
   {
     slots[i].write_buf= static_cast<byte*>
@@ -87,7 +87,7 @@ bool buf_dblwr_t::create()
     return true;
 
   mtr_t mtr;
-  const ulint size= block_size();
+  const ulint size= block_size;
 
 start_again:
   mtr.start();
@@ -252,7 +252,7 @@ loads the pages from double write buffer into memory.
 dberr_t buf_dblwr_t::init_or_load_pages(pfs_os_file_t file, const char *path)
 {
   ut_ad(this == &buf_dblwr);
-  const uint32_t size= block_size();
+  const uint32_t size= block_size;
 
   /* We do the file i/o past the buffer pool */
   byte *read_buf= static_cast<byte*>(aligned_malloc(srv_page_size,
@@ -284,7 +284,6 @@ func_exit:
   init(TRX_SYS_DOUBLEWRITE + read_buf);
 
   const bool upgrade_to_innodb_file_per_table=
-    !srv_read_only_mode &&
     mach_read_from_4(TRX_SYS_DOUBLEWRITE_SPACE_ID_STORED +
                      TRX_SYS_DOUBLEWRITE + read_buf) !=
     TRX_SYS_DOUBLEWRITE_SPACE_ID_STORED_N;
@@ -356,9 +355,6 @@ void buf_dblwr_t::recover()
   ut_ad(log_sys.last_checkpoint_lsn);
   if (!is_created())
     return;
-  const lsn_t max_lsn{log_sys.get_lsn()};
-  ut_ad(recv_sys.scanned_lsn == max_lsn);
-  ut_ad(recv_sys.scanned_lsn >= recv_sys.lsn);
 
   uint32_t page_no_dblwr= 0;
   byte *read_buf= static_cast<byte*>(aligned_malloc(3 * srv_page_size,
@@ -368,15 +364,25 @@ void buf_dblwr_t::recover()
   for (recv_dblwr_t::list::iterator i= recv_sys.dblwr.pages.begin();
        i != recv_sys.dblwr.pages.end(); ++i, ++page_no_dblwr)
   {
-    const page_t *const page= *i;
+    byte *page= *i;
     const uint32_t page_no= page_get_page_no(page);
+    if (!page_no) /* recovered via recv_dblwr_t::restore_first_page() */
+      continue;
+
     const lsn_t lsn= mach_read_from_8(page + FIL_PAGE_LSN);
-    if (log_sys.last_checkpoint_lsn > lsn || lsn > recv_sys.lsn)
-      /* Pages written before or after the recovery range are not usable. */
+    if (log_sys.last_checkpoint_lsn > lsn)
+      /* Pages written before the checkpoint are not useful for recovery. */
       continue;
     const uint32_t space_id= page_get_space_id(page);
     const page_id_t page_id(space_id, page_no);
 
+    if (recv_sys.scanned_lsn < lsn)
+    {
+      ib::info() << "Ignoring a doublewrite copy of page " << page_id
+                 << " with future log sequence number " << lsn;
+      continue;
+    }
+
     fil_space_t *space= fil_space_t::get(space_id);
 
     if (!space)
@@ -388,14 +394,10 @@ void buf_dblwr_t::recover()
       /* Do not report the warning for undo tablespaces, because they
       can be truncated in place. */
       if (!srv_is_undo_tablespace(space_id))
-        sql_print_warning("InnoDB: A copy of page "
-                          "[page id: space=" UINT32PF
-                          ", page number=" UINT32PF "]"
-                         " in the doublewrite buffer slot " UINT32PF
-                          " is beyond the end of %s (" UINT32PF " pages)",
-                          page_id.space(), page_id.page_no(),
-                          page_no_dblwr, space->chain.start->name,
-                          space->size);
+        ib::warn() << "A copy of page " << page_no
+                   << " in the doublewrite buffer slot " << page_no_dblwr
+                   << " is beyond the end of " << space->chain.start->name
+                   << " (" << space->size << " pages)";
 next_page:
       space->release();
       continue;
@@ -414,48 +416,41 @@ next_page:
                             physical_size, read_buf);
 
     if (UNIV_UNLIKELY(fio.err != DB_SUCCESS))
-      sql_print_warning("InnoDB: Double write buffer recovery: "
-                        "[page id: space=" UINT32PF
-                        ", page number=" UINT32PF "]"
-                        " ('%s') read failed with error: %s",
-                        page_id.space(), page_id.page_no(), fio.node->name,
-                        ut_strerr(fio.err));
-    else if (buf_is_zeroes(span<const byte>(read_buf, physical_size)))
+    {
+       ib::warn() << "Double write buffer recovery: " << page_id
+                  << " ('" << space->chain.start->name
+                  << "') read failed with error: " << fio.err;
+       continue;
+    }
+
+    if (buf_is_zeroes(span<const byte>(read_buf, physical_size)))
     {
       /* We will check if the copy in the doublewrite buffer is
       valid. If not, we will ignore this page (there should be redo
       log records to initialize it). */
     }
-    else if (recv_sys.dblwr.validate_page(page_id, max_lsn, space,
-                                          read_buf, buf))
+    else if (recv_sys.dblwr.validate_page(page_id, read_buf, space, buf))
       goto next_page;
     else
       /* We intentionally skip this message for all-zero pages. */
-      sql_print_information("InnoDB: Trying to recover page "
-                            "[page id: space=" UINT32PF
-                            ", page number=" UINT32PF "]"
-                            " from the doublewrite buffer.",
-                            page_id.space(), page_id.page_no());
+      ib::info() << "Trying to recover page " << page_id
+                 << " from the doublewrite buffer.";
 
-    if (const byte *page=
-        recv_sys.dblwr.find_page(page_id, max_lsn, space, buf))
-    {
-      /* Write the good page from the doublewrite buffer to the intended
-      position. */
-      space->reacquire();
-      fio= space->io(IORequestWrite,
-                     os_offset_t{page_id.page_no()} * physical_size,
-                     physical_size, const_cast<byte*>(page));
+    page= recv_sys.dblwr.find_page(page_id, space, buf);
 
-      if (fio.err == DB_SUCCESS)
-        sql_print_information("InnoDB: Recovered page "
-                              "[page id: space=" UINT32PF
-                              ", page number=" UINT32PF "]"
-                              " to '%s' from the doublewrite buffer.",
-                              page_id.space(), page_id.page_no(),
-                              fio.node->name);
-    }
+    if (!page)
+      goto next_page;
 
+    /* Write the good page from the doublewrite buffer to the intended
+    position. */
+    space->reacquire();
+    fio= space->io(IORequestWrite,
+                   os_offset_t{page_id.page_no()} * physical_size,
+                   physical_size, page);
+
+    if (fio.err == DB_SUCCESS)
+      ib::info() << "Recovered page " << page_id << " to '" << fio.node->name
+                 << "' from the doublewrite buffer.";
     goto next_page;
   }
 
@@ -494,7 +489,6 @@ void buf_dblwr_t::write_completed()
   mysql_mutex_lock(&mutex);
 
   ut_ad(is_created());
-  ut_ad(srv_use_doublewrite_buf);
   ut_ad(batch_running);
   slot *flush_slot= active_slot == &slots[0] ? &slots[1] : &slots[0];
   ut_ad(flush_slot->reserved);
@@ -580,7 +574,7 @@ static void buf_dblwr_check_block(const buf_page_t *bpage)
 bool buf_dblwr_t::flush_buffered_writes(const ulint size)
 {
   mysql_mutex_assert_owner(&mutex);
-  ut_ad(size == block_size());
+  ut_ad(size == block_size);
 
   for (;;)
   {
@@ -653,7 +647,6 @@ static void *get_frame(const IORequest &request)
 void buf_dblwr_t::flush_buffered_writes_completed(const IORequest &request)
 {
   ut_ad(this == &buf_dblwr);
-  ut_ad(srv_use_doublewrite_buf);
   ut_ad(is_created());
   ut_ad(!srv_read_only_mode);
   ut_ad(!request.bpage);
@@ -676,8 +669,14 @@ void buf_dblwr_t::flush_buffered_writes_completed(const IORequest &request)
   pages_written+= flush_slot->first_free;
   mysql_mutex_unlock(&mutex);
 
-  /* Now flush the doublewrite buffer data to disk */
-  fil_system.sys_space->flush<false>();
+  /* Make the doublewrite durable. Note: The doublewrite buffer is
+  always in the first file of the system tablespace. We will not
+  bother about fil_system.unflushed_spaces, which can result in a
+  redundant call during fil_flush_file_spaces() in
+  log_checkpoint(). Writes to the system tablespace should be rare,
+  except when executing DDL or using the non-default settings
+  innodb_file_per_table=OFF or innodb_undo_tablespaces=0. */
+  os_file_flush(request.node->handle);
 
   /* The writes have been flushed to disk now and in recovery we will
   find them in the doublewrite buffer blocks. Next, write the data pages. */
@@ -720,17 +719,18 @@ posted, and also when we may have to wait for a page latch!
 Otherwise a deadlock of threads can occur. */
 void buf_dblwr_t::flush_buffered_writes()
 {
-  if (!is_created() || !srv_use_doublewrite_buf)
+  mysql_mutex_lock(&mutex);
+
+  if (!in_use() && active_slot->first_free == 0)
   {
+    mysql_mutex_unlock(&mutex);
     fil_flush_file_spaces();
     return;
   }
 
   ut_ad(!srv_read_only_mode);
-  const ulint size= block_size();
 
-  mysql_mutex_lock(&mutex);
-  if (!flush_buffered_writes(size))
+  if (!flush_buffered_writes(block_size))
     mysql_mutex_unlock(&mutex);
 }
 
@@ -740,8 +740,6 @@ flush_buffered_writes() will be invoked to make space.
 @param size       payload size in bytes */
 void buf_dblwr_t::add_to_batch(const IORequest &request, size_t size)
 {
-  ut_ad(request.is_async());
-  ut_ad(request.is_write());
   ut_ad(request.bpage);
   ut_ad(request.bpage->in_file());
   ut_ad(request.node);
@@ -750,7 +748,7 @@ void buf_dblwr_t::add_to_batch(const IORequest &request, size_t size)
   ut_ad(request.node->space->referenced());
   ut_ad(!srv_read_only_mode);
 
-  const ulint buf_size= 2 * block_size();
+  const ulint buf_size= 2 * block_size;
 
   mysql_mutex_lock(&mutex);
 
@@ -779,7 +777,7 @@ void buf_dblwr_t::add_to_batch(const IORequest &request, size_t size)
   ut_ad(active_slot->reserved == active_slot->first_free);
   ut_ad(active_slot->reserved < buf_size);
   new (active_slot->buf_block_arr + active_slot->first_free++)
-    element{request, size};
+    element{request.doublewritten(), size};
   active_slot->reserved= active_slot->first_free;
 
   if (active_slot->first_free != buf_size ||
