@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2013, 2022, MariaDB Corporation.
+Copyright (c) 2013, 2023, MariaDB Corporation.
 Copyright (c) 2013, 2014, Fusion-io
 
 This program is free software; you can redistribute it and/or modify it under
@@ -350,9 +350,9 @@ void buf_page_write_complete(const IORequest &request, bool error)
   else
   {
     bpage->write_complete(persistent, error, state);
-    if (state < buf_page_t::WRITE_FIX_REINIT &&
-        request.node->space->use_doublewrite())
+    if (request.is_doublewritten())
     {
+      ut_ad(state < buf_page_t::WRITE_FIX_REINIT);
       ut_ad(persistent);
       buf_dblwr.write_completed();
     }
@@ -1069,7 +1069,6 @@ static ulint buf_flush_try_neighbors(fil_space_t *space,
       {
         ut_ad(bpage == b);
         bpage= nullptr;
-        ut_ad(!buf_pool.watch_is_sentinel(*b));
         ut_ad(b->oldest_modification() > 1);
       flush:
         if (b->flush(space))
@@ -1083,7 +1082,6 @@ static ulint buf_flush_try_neighbors(fil_space_t *space,
       else if (b->oldest_modification() > 1 && b->lock.u_lock_try(true))
       {
         /* For the buf_pool.watch[] sentinels, oldest_modification() == 0 */
-        ut_ad(!buf_pool.watch_is_sentinel(*b));
         if (b->oldest_modification() < 2)
           b->lock.u_unlock(true);
         else
@@ -1120,7 +1118,7 @@ static ulint buf_free_from_unzip_LRU_list_batch()
 	buf_block_t*	block = UT_LIST_GET_LAST(buf_pool.unzip_LRU);
 
 	while (block
-	       && UT_LIST_GET_LEN(buf_pool.free) < buf_pool.LRU_scan_depth
+	       && UT_LIST_GET_LEN(buf_pool.free) < srv_LRU_scan_depth
 	       && UT_LIST_GET_LEN(buf_pool.unzip_LRU)
 	       > UT_LIST_GET_LEN(buf_pool.LRU) / 10) {
 
@@ -1197,7 +1195,7 @@ static void buf_flush_discard_page(buf_page_t *bpage)
 
   ut_d(const auto state= bpage->state());
   ut_ad(state == buf_page_t::FREED || state == buf_page_t::UNFIXED ||
-        state == buf_page_t::IBUF_EXIST || state == buf_page_t::REINIT);
+        state == buf_page_t::REINIT);
   bpage->lock.u_unlock(true);
   buf_LRU_free_page(bpage, true);
 }
@@ -1209,33 +1207,22 @@ and move clean blocks to buf_pool.free.
 static void buf_flush_LRU_list_batch(ulint max, flush_counters_t *n)
 {
   ulint scanned= 0;
+  ulint free_limit= srv_LRU_scan_depth;
+
   mysql_mutex_assert_owner(&buf_pool.mutex);
-  ulint free_limit{buf_pool.LRU_scan_depth};
   if (buf_pool.withdraw_target && buf_pool.is_shrinking())
     free_limit+= buf_pool.withdraw_target - UT_LIST_GET_LEN(buf_pool.withdraw);
 
   const auto neighbors= UT_LIST_GET_LEN(buf_pool.LRU) < BUF_LRU_OLD_MIN_LEN
-    ? 0 : buf_pool.flush_neighbors;
+    ? 0 : srv_flush_neighbors;
   fil_space_t *space= nullptr;
   uint32_t last_space_id= FIL_NULL;
   static_assert(FIL_NULL > SRV_TMP_SPACE_ID, "consistency");
   static_assert(FIL_NULL > SRV_SPACE_ID_UPPER_BOUND, "consistency");
 
-  /* BUF_LRU_MIN_LEN (256) is too high value for low buffer pool(BP) size. For
-  example, for BP size lower than 80M and 16 K page size, the limit is more than
-  5% of total BP and for lowest BP 5M, it is 80% of the BP. Non-data objects
-  like explicit locks could occupy part of the BP pool reducing the pages
-  available for LRU. If LRU reaches minimum limit and if no free pages are
-  available, server would hang with page cleaner not able to free any more
-  pages. To avoid such hang, we adjust the LRU limit lower than the limit for
-  data objects as checked in buf_LRU_check_size_of_non_data_objects() i.e. one
-  page less than 5% of BP. */
-  size_t pool_limit= buf_pool.curr_size / 20 - 1;
-  auto buf_lru_min_len= std::min<size_t>(pool_limit, BUF_LRU_MIN_LEN);
-
   for (buf_page_t *bpage= UT_LIST_GET_LAST(buf_pool.LRU);
        bpage &&
-       ((UT_LIST_GET_LEN(buf_pool.LRU) > buf_lru_min_len &&
+       ((UT_LIST_GET_LEN(buf_pool.LRU) > BUF_LRU_MIN_LEN &&
          UT_LIST_GET_LEN(buf_pool.free) < free_limit) ||
         recv_recovery_is_on());
        ++scanned, bpage= buf_pool.lru_hp.get())
@@ -1391,7 +1378,7 @@ static ulint buf_do_flush_list_batch(ulint max_n, lsn_t lsn)
   mysql_mutex_assert_owner(&buf_pool.flush_list_mutex);
 
   const auto neighbors= UT_LIST_GET_LEN(buf_pool.LRU) < BUF_LRU_OLD_MIN_LEN
-    ? 0 : buf_pool.flush_neighbors;
+    ? 0 : srv_flush_neighbors;
   fil_space_t *space= nullptr;
   uint32_t last_space_id= FIL_NULL;
   static_assert(FIL_NULL > SRV_TMP_SPACE_ID, "consistency");
@@ -1595,7 +1582,7 @@ static ulint buf_flush_list(ulint max_n= ULINT_UNDEFINED,
 bool buf_flush_list_space(fil_space_t *space, ulint *n_flushed)
 {
   const auto space_id= space->id;
-  ut_ad(space_id < SRV_SPACE_ID_UPPER_BOUND);
+  ut_ad(space_id <= SRV_SPACE_ID_UPPER_BOUND);
 
   bool may_have_skipped= false;
   ulint max_n_flush= srv_io_capacity;
@@ -1765,7 +1752,7 @@ inline void log_t::write_checkpoint(lsn_t end_lsn) noexcept
   static_assert(CPU_LEVEL1_DCACHE_LINESIZE >= 64, "efficiency");
   static_assert(CPU_LEVEL1_DCACHE_LINESIZE <= 4096, "compatibility");
   byte* c= my_assume_aligned<CPU_LEVEL1_DCACHE_LINESIZE>
-    (is_mmap() ? buf + offset : checkpoint_buf);
+    (is_pmem() ? buf + offset : checkpoint_buf);
   memset_aligned<CPU_LEVEL1_DCACHE_LINESIZE>(c, 0, CPU_LEVEL1_DCACHE_LINESIZE);
   mach_write_to_8(my_assume_aligned<8>(c), next_checkpoint_lsn);
   mach_write_to_8(my_assume_aligned<8>(c + 8), end_lsn);
@@ -1774,9 +1761,8 @@ inline void log_t::write_checkpoint(lsn_t end_lsn) noexcept
   lsn_t resizing;
 
 #ifdef HAVE_PMEM
-  if (is_mmap())
+  if (is_pmem())
   {
-    ut_ad(!is_opened());
     resizing= resize_lsn.load(std::memory_order_relaxed);
 
     if (resizing > 1 && resizing <= next_checkpoint_lsn)
@@ -1790,27 +1776,24 @@ inline void log_t::write_checkpoint(lsn_t end_lsn) noexcept
   else
 #endif
   {
-    ut_ad(!is_mmap());
     ut_ad(!checkpoint_pending);
     checkpoint_pending= true;
     latch.wr_unlock();
     log_write_and_flush_prepare();
     resizing= resize_lsn.load(std::memory_order_relaxed);
-    ut_ad(ut_is_2pow(write_size));
-    ut_ad(write_size >= 512);
-    ut_ad(write_size <= 4096);
-    log.write(offset, {c, write_size});
+    /* FIXME: issue an asynchronous write */
+    log.write(offset, {c, get_block_size()});
     if (resizing > 1 && resizing <= next_checkpoint_lsn)
     {
-      resize_log.write(CHECKPOINT_1, {c, write_size});
       byte *buf= static_cast<byte*>(aligned_malloc(4096, 4096));
       memset_aligned<4096>(buf, 0, 4096);
       header_write(buf, resizing, is_encrypted());
       resize_log.write(0, {buf, 4096});
       aligned_free(buf);
+      resize_log.write(CHECKPOINT_1, {c, get_block_size()});
     }
 
-    if (srv_file_flush_method != SRV_O_DSYNC)
+    if (!log_write_through)
       ut_a(log.flush());
     latch.wr_lock(SRW_LOCK_CALL);
     ut_ad(checkpoint_pending);
@@ -1838,28 +1821,24 @@ inline void log_t::write_checkpoint(lsn_t end_lsn) noexcept
 
   if (resizing > 1 && resizing <= checkpoint_lsn)
   {
-    ut_ad(is_mmap() == !resize_flush_buf);
+    ut_ad(is_pmem() == !resize_flush_buf);
 
-    if (!is_mmap())
+    if (!is_pmem())
     {
-      if (srv_file_flush_method != SRV_O_DSYNC)
+      if (!log_write_through)
         ut_a(resize_log.flush());
       IF_WIN(log.close(),);
     }
 
     if (resize_rename())
     {
-      /* Resizing failed. Discard the ib_logfile101. */
+      /* Resizing failed. Discard the log_sys.resize_log. */
 #ifdef HAVE_PMEM
-      if (is_mmap())
-      {
-        ut_ad(!is_opened());
+      if (is_pmem())
         my_munmap(resize_buf, resize_target);
-      }
       else
 #endif
       {
-        ut_ad(!is_mmap());
         ut_free_dodump(resize_buf, buf_size);
         ut_free_dodump(resize_flush_buf, buf_size);
 #ifdef _WIN32
@@ -1867,7 +1846,7 @@ inline void log_t::write_checkpoint(lsn_t end_lsn) noexcept
         bool success;
         log.m_file=
           os_file_create_func(get_log_file_path().c_str(), OS_FILE_OPEN,
-                              OS_LOG_FILE, false, &success);
+                              OS_FILE_NORMAL, OS_LOG_FILE, false, &success);
         ut_a(success);
         ut_a(log.is_opened());
 #endif
@@ -1877,9 +1856,8 @@ inline void log_t::write_checkpoint(lsn_t end_lsn) noexcept
     {
       /* Adopt the resized log. */
 #ifdef HAVE_PMEM
-      if (is_mmap())
+      if (is_pmem())
       {
-        ut_ad(!is_opened());
         my_munmap(buf, file_size);
         buf= resize_buf;
         set_buf_free(START_OFFSET + (get_lsn() - resizing));
@@ -1887,7 +1865,6 @@ inline void log_t::write_checkpoint(lsn_t end_lsn) noexcept
       else
 #endif
       {
-        ut_ad(!is_mmap());
         IF_WIN(,log.close());
         std::swap(log, resize_log);
         ut_free_dodump(buf, buf_size);
@@ -1990,13 +1967,7 @@ static bool log_checkpoint()
   if (recv_recovery_is_on())
     recv_sys.apply(true);
 
-  switch (srv_file_flush_method) {
-  case SRV_NOSYNC:
-  case SRV_O_DIRECT_NO_FSYNC:
-    break;
-  default:
-    fil_flush_file_spaces();
-  }
+  fil_flush_file_spaces();
 
   log_sys.latch.wr_lock(SRW_LOCK_CALL);
   const lsn_t end_lsn= log_sys.get_lsn();
@@ -2139,90 +2110,60 @@ ATTRIBUTE_COLD void buf_flush_ahead(lsn_t lsn, bool furious)
 /** Conduct checkpoint-related flushing for innodb_flush_sync=ON,
 and try to initiate checkpoints until the target is met.
 @param lsn   minimum value of buf_pool.get_oldest_modification(LSN_MAX) */
-ATTRIBUTE_COLD ATTRIBUTE_NOINLINE
-static void buf_flush_sync_for_checkpoint(lsn_t lsn)
+ATTRIBUTE_COLD static void buf_flush_sync_for_checkpoint(lsn_t lsn)
 {
   ut_ad(!srv_read_only_mode);
   mysql_mutex_assert_not_owner(&buf_pool.flush_list_mutex);
 
-  /* During furious flush, we need to keep generating free pages. Otherwise
-  concurrent mtrs could be blocked holding latches for the pages to be flushed
-  causing deadlock in rare occasion.
-
-  Ideally we should be acquiring buffer pool mutex for the check but it is more
-  expensive and we are not using the mutex while calling need_LRU_eviction() as
-  of today. It is a quick and dirty read of the LRU and free list length.
-  Atomic read of try_LRU_scan should eventually let us do the eviction.
-  Correcting the inaccuracy would need more consideration to avoid any possible
-  performance regression. */
-  if (buf_pool.need_LRU_eviction())
+  for (;;)
   {
-    mysql_mutex_lock(&buf_pool.flush_list_mutex);
-    buf_pool.page_cleaner_set_idle(false);
-    buf_pool.n_flush_inc();
-    mysql_mutex_unlock(&buf_pool.flush_list_mutex);
+    if (ulint n_flushed= buf_flush_list(srv_max_io_capacity, lsn))
+    {
+      MONITOR_INC_VALUE_CUMULATIVE(MONITOR_FLUSH_SYNC_TOTAL_PAGE,
+                                   MONITOR_FLUSH_SYNC_COUNT,
+                                   MONITOR_FLUSH_SYNC_PAGES, n_flushed);
+    }
 
-    mysql_mutex_lock(&buf_pool.mutex);
-    /* Confirm that eviction is needed after acquiring buffer pool mutex. */
-    if (buf_pool.need_LRU_eviction())
-      /* We intend to only evict pages keeping maximum flush bandwidth for
-      flush list pages advancing checkpoint. However, if the LRU tail is full
-      of dirty pages, we might need some flushing. */
-      std::ignore= buf_flush_LRU(srv_io_capacity);
-    mysql_mutex_unlock(&buf_pool.mutex);
-
-    mysql_mutex_lock(&buf_pool.flush_list_mutex);
-    buf_pool.n_flush_dec();
-    mysql_mutex_unlock(&buf_pool.flush_list_mutex);
-  }
-
-  if (ulint n_flushed= buf_flush_list(srv_max_io_capacity, lsn))
-  {
-    MONITOR_INC_VALUE_CUMULATIVE(MONITOR_FLUSH_SYNC_TOTAL_PAGE,
-                                 MONITOR_FLUSH_SYNC_COUNT,
-                                 MONITOR_FLUSH_SYNC_PAGES, n_flushed);
-  }
-
-  switch (srv_file_flush_method) {
-  case SRV_NOSYNC:
-  case SRV_O_DIRECT_NO_FSYNC:
-    break;
-  default:
     fil_flush_file_spaces();
-  }
 
-  log_sys.latch.wr_lock(SRW_LOCK_CALL);
-  const lsn_t newest_lsn= log_sys.get_lsn();
-  mysql_mutex_lock(&buf_pool.flush_list_mutex);
-  lsn_t measure= buf_pool.get_oldest_modification(0);
-  const lsn_t checkpoint_lsn= measure ? measure : newest_lsn;
-
-  if (!recv_recovery_is_on() &&
-      checkpoint_lsn > log_sys.last_checkpoint_lsn + SIZE_OF_FILE_CHECKPOINT)
-  {
-    mysql_mutex_unlock(&buf_pool.flush_list_mutex);
-    log_checkpoint_low(checkpoint_lsn, newest_lsn);
+    log_sys.latch.wr_lock(SRW_LOCK_CALL);
+    const lsn_t newest_lsn= log_sys.get_lsn();
     mysql_mutex_lock(&buf_pool.flush_list_mutex);
-    measure= buf_pool.get_oldest_modification(LSN_MAX);
+    lsn_t measure= buf_pool.get_oldest_modification(0);
+    const lsn_t checkpoint_lsn= measure ? measure : newest_lsn;
+
+    if (!recv_recovery_is_on() &&
+        checkpoint_lsn > log_sys.last_checkpoint_lsn + SIZE_OF_FILE_CHECKPOINT)
+    {
+      mysql_mutex_unlock(&buf_pool.flush_list_mutex);
+      log_checkpoint_low(checkpoint_lsn, newest_lsn);
+      mysql_mutex_lock(&buf_pool.flush_list_mutex);
+      measure= buf_pool.get_oldest_modification(LSN_MAX);
+    }
+    else
+    {
+      log_sys.latch.wr_unlock();
+      if (!measure)
+        measure= LSN_MAX;
+    }
+
+    /* After attempting log checkpoint, check if we have reached our target. */
+    const lsn_t target= buf_flush_sync_lsn;
+
+    if (measure >= target)
+      buf_flush_sync_lsn= 0;
+    else if (measure >= buf_flush_async_lsn)
+      buf_flush_async_lsn= 0;
+
+    /* wake up buf_flush_wait() */
+    pthread_cond_broadcast(&buf_pool.done_flush_list);
+    mysql_mutex_unlock(&buf_pool.flush_list_mutex);
+
+    lsn= std::max(lsn, target);
+
+    if (measure >= lsn)
+      return;
   }
-  else
-  {
-    log_sys.latch.wr_unlock();
-    if (!measure)
-      measure= LSN_MAX;
-  }
-
-  /* After attempting log checkpoint, check if we have reached our target. */
-  const lsn_t target= buf_flush_sync_lsn;
-
-  if (measure >= target)
-    buf_flush_sync_lsn= 0;
-  else if (measure >= buf_flush_async_lsn)
-    buf_flush_async_lsn= 0;
-
-  /* wake up buf_flush_wait() */
-  pthread_cond_broadcast(&buf_pool.done_flush_list);
-  mysql_mutex_unlock(&buf_pool.flush_list_mutex);
 }
 
 /** Check if the adpative flushing threshold is recommended based on
@@ -2316,7 +2257,7 @@ func_exit:
 
 	sum_pages += last_pages_in;
 
-	const ulint time_elapsed = std::max<ulint>(ulint(curr_time - prev_time), 1);
+	const ulint time_elapsed = std::max<ulint>(curr_time - prev_time, 1);
 
 	/* We update our variables every innodb_flushing_avg_loops
 	iterations to smooth out transition in workload. */
@@ -2409,7 +2350,7 @@ bool buf_pool_t::need_LRU_eviction() const
   for buf_flush_page_cleaner() to evict some blocks */
   return UNIV_UNLIKELY(!try_LRU_scan ||
                        (UT_LIST_GET_LEN(LRU) > BUF_LRU_MIN_LEN &&
-                        UT_LIST_GET_LEN(free) < LRU_scan_depth / 2));
+                        UT_LIST_GET_LEN(free) < srv_LRU_scan_depth / 2));
 }
 
 #if defined __aarch64__&&defined __GNUC__&&__GNUC__==4&&!defined __clang__
@@ -2436,10 +2377,6 @@ static void buf_flush_page_cleaner()
 
   for (;;)
   {
-    DBUG_EXECUTE_IF("ib_page_cleaner_sleep",
-    {
-      std::this_thread::sleep_for(std::chrono::seconds(1));
-    });
     lsn_limit= buf_flush_sync_lsn;
 
     if (UNIV_UNLIKELY(lsn_limit != 0) && UNIV_LIKELY(srv_flush_sync))
@@ -2702,12 +2639,12 @@ ATTRIBUTE_COLD void buf_flush_page_cleaner_init()
 /** Flush the buffer pool on shutdown. */
 ATTRIBUTE_COLD void buf_flush_buffer_pool()
 {
+  ut_ad(!os_aio_pending_reads());
   ut_ad(!buf_page_cleaner_is_active);
   ut_ad(!buf_flush_sync_lsn);
 
   service_manager_extend_timeout(INNODB_EXTEND_TIMEOUT_INTERVAL,
                                  "Waiting to flush the buffer pool");
-  os_aio_wait_until_no_pending_reads(false);
 
   mysql_mutex_lock(&buf_pool.flush_list_mutex);
 
