@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2021, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2014, 2022, MariaDB Corporation.
+Copyright (c) 2014, 2023, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -59,34 +59,27 @@ Created 10/25/1995 Heikki Tuuri
 #include "bzlib.h"
 #include "snappy-c.h"
 
-ATTRIBUTE_COLD bool fil_space_t::set_corrupted() const noexcept
+ATTRIBUTE_COLD void fil_space_t::set_corrupted() const
 {
   if (!is_stopping() && !is_corrupted.test_and_set())
-  {
     sql_print_error("InnoDB: File '%s' is corrupted", chain.start->name);
-    return true;
-  }
-  return false;
 }
 
 /** Try to close a file to adhere to the innodb_open_files limit.
-@param ignore_space Ignore the tablespace which is acquired by caller
 @param print_info   whether to diagnose why a file cannot be closed
 @return whether a file was closed */
-bool fil_space_t::try_to_close(fil_space_t *ignore_space, bool print_info)
+bool fil_space_t::try_to_close(bool print_info)
 {
   mysql_mutex_assert_owner(&fil_system.mutex);
   for (fil_space_t &space : fil_system.space_list)
   {
-    if (&space == ignore_space)
-      continue;
     switch (space.purpose) {
     case FIL_TYPE_TEMPORARY:
       continue;
     case FIL_TYPE_IMPORT:
       break;
     case FIL_TYPE_TABLESPACE:
-      if (space.id == TRX_SYS_SPACE || srv_is_undo_tablespace(space.id))
+      if (is_predefined_tablespace(space.id))
         continue;
     }
 
@@ -331,7 +324,7 @@ fil_node_t* fil_space_t::add(const char* name, pfs_os_file_t handle,
 		clear_closing();
 		if (++fil_system.n_open >= srv_max_n_open_files) {
 			reacquire();
-			try_to_close(this, true);
+			try_to_close(true);
 			release();
 		}
 	}
@@ -339,14 +332,11 @@ fil_node_t* fil_space_t::add(const char* name, pfs_os_file_t handle,
 	return node;
 }
 
-__attribute__((warn_unused_result, nonnull(1)))
+__attribute__((warn_unused_result, nonnull))
 /** Open a tablespace file.
 @param node  data file
-@param page  first page of the tablespace, or nullptr
-@param no_lsn  whether to skip the FIL_PAGE_LSN check
 @return whether the file was successfully opened */
-static bool fil_node_open_file_low(fil_node_t *node, const byte *page,
-                                   bool no_lsn)
+static bool fil_node_open_file_low(fil_node_t *node)
 {
   ut_ad(!node->is_open());
   ut_ad(node->space->is_closing());
@@ -372,7 +362,7 @@ static bool fil_node_open_file_low(fil_node_t *node, const byte *page,
     node->handle= os_file_create(innodb_data_file_key, node->name,
                                  node->is_raw_disk
                                  ? OS_FILE_OPEN_RAW : OS_FILE_OPEN,
-                                 type,
+                                 OS_FILE_AIO, type,
                                  srv_read_only_mode, &success);
 
     if (success && node->is_open())
@@ -391,7 +381,7 @@ static bool fil_node_open_file_low(fil_node_t *node, const byte *page,
 
     /* The following call prints an error message */
     if (os_file_get_last_error(true) == EMFILE + 100 &&
-        fil_space_t::try_to_close(nullptr, true))
+        fil_space_t::try_to_close(true))
       continue;
 
     ib::warn() << "Cannot open '" << node->name << "'.";
@@ -402,7 +392,7 @@ static bool fil_node_open_file_low(fil_node_t *node, const byte *page,
   bool comp_algo_invalid = false;
 
   if (node->size);
-  else if (!node->read_page0(page, no_lsn) ||
+  else if (!node->read_page0() ||
             // validate compression algorithm for full crc32 format
             (node->space->full_crc32() &&
              (comp_algo_invalid = !fil_comp_algo_loaded(comp_algo))))
@@ -433,10 +423,8 @@ static bool fil_node_open_file_low(fil_node_t *node, const byte *page,
 
 /** Open a tablespace file.
 @param node  data file
-@param page  first page of the tablespace, or nullptr
-@param no_lsn  whether to skip the FIL_PAGE_LSN check
 @return whether the file was successfully opened */
-static bool fil_node_open_file(fil_node_t *node, const byte *page, bool no_lsn)
+static bool fil_node_open_file(fil_node_t *node)
 {
   mysql_mutex_assert_owner(&fil_system.mutex);
   ut_ad(!node->is_open());
@@ -451,7 +439,7 @@ static bool fil_node_open_file(fil_node_t *node, const byte *page, bool no_lsn)
 
   for (ulint count= 0; fil_system.n_open >= srv_max_n_open_files; count++)
   {
-    if (fil_space_t::try_to_close(nullptr, count > 1))
+    if (fil_space_t::try_to_close(count > 1))
       count= 0;
     else if (count >= 2)
     {
@@ -475,7 +463,7 @@ static bool fil_node_open_file(fil_node_t *node, const byte *page, bool no_lsn)
 
   /* The node can be opened beween releasing and acquiring fil_system.mutex
   in the above code */
-  return node->is_open() || fil_node_open_file_low(node, page, no_lsn);
+  return node->is_open() || fil_node_open_file_low(node);
 }
 
 /** Close the file handle. */
@@ -529,6 +517,9 @@ void fil_space_t::flush_low()
       break;
   }
 
+  if (fil_system.is_write_through())
+    goto skip_flush;
+
   fil_n_pending_tablespace_flushes++;
   for (fil_node_t *node= UT_LIST_GET_FIRST(chain); node;
        node= UT_LIST_GET_NEXT(chain, node))
@@ -553,8 +544,9 @@ void fil_space_t::flush_low()
     mysql_mutex_unlock(&fil_system.mutex);
   }
 
-  clear_flush();
   fil_n_pending_tablespace_flushes--;
+skip_flush:
+  clear_flush();
 }
 
 /** Try to extend a tablespace.
@@ -686,8 +678,7 @@ ATTRIBUTE_COLD bool fil_space_t::prepare_acquired()
   ut_ad(!id || purpose == FIL_TYPE_TEMPORARY ||
         node == UT_LIST_GET_FIRST(chain));
 
-  const bool is_open= node &&
-    (node->is_open() || fil_node_open_file(node, nullptr, false));
+  const bool is_open= node && (node->is_open() || fil_node_open_file(node));
 
   if (!is_open)
     release();
@@ -781,7 +772,6 @@ inline pfs_os_file_t fil_node_t::close_to_free(bool detach_handle)
   {
     if (space->is_in_unflushed_spaces)
     {
-      ut_ad(srv_file_flush_method != SRV_O_DIRECT_NO_FSYNC);
       space->is_in_unflushed_spaces= false;
       fil_system.unflushed_spaces.remove(*space);
     }
@@ -814,7 +804,6 @@ pfs_os_file_t fil_system_t::detach(fil_space_t *space, bool detach_handle)
 
   if (space->is_in_unflushed_spaces)
   {
-    ut_ad(srv_file_flush_method != SRV_O_DIRECT_NO_FSYNC);
     space->is_in_unflushed_spaces= false;
     unflushed_spaces.remove(*space);
   }
@@ -1114,10 +1103,8 @@ bool fil_assign_new_space_id(uint32_t *space_id)
 }
 
 /** Read the first page of a data file.
-@param dpage   copy of a first page, from the doublewrite buffer, or nullptr
-@param no_lsn  whether to skip the FIL_PAGE_LSN check
 @return whether the page was found valid */
-bool fil_space_t::read_page0(const byte *dpage, bool no_lsn) noexcept
+bool fil_space_t::read_page0()
 {
   ut_ad(fil_system.is_initialised());
   mysql_mutex_assert_owner(&fil_system.mutex);
@@ -1134,26 +1121,32 @@ bool fil_space_t::read_page0(const byte *dpage, bool no_lsn) noexcept
     ut_ad("this should not happen" == 0);
     return false;
   }
-  const bool ok= node->is_open() || fil_node_open_file(node, dpage, no_lsn);
+  const bool ok= node->is_open() || fil_node_open_file(node);
   release();
   return ok;
 }
 
+/** Look up a tablespace and ensure that its first page has been validated. */
+static fil_space_t *fil_space_get_space(uint32_t id)
+{
+  if (fil_space_t *space= fil_space_get_by_id(id))
+    if (space->read_page0())
+      return space;
+  return nullptr;
+}
+
 void fil_space_set_recv_size_and_flags(uint32_t id, uint32_t size,
                                        uint32_t flags)
 {
   ut_ad(id < SRV_SPACE_ID_UPPER_BOUND);
-  mysql_mutex_assert_owner(&recv_sys.mutex);
   mysql_mutex_lock(&fil_system.mutex);
-  if (fil_space_t *space= fil_space_get_by_id(id))
-    if (space->read_page0(recv_sys.dblwr.find_page(page_id_t(id, 0), LSN_MAX),
-                          true))
-    {
-      if (size)
-        space->recv_size= size;
-      if (flags != FSP_FLAGS_FCRC32_MASK_MARKER)
-        space->flags= flags;
-    }
+  if (fil_space_t *space= fil_space_get_space(id))
+  {
+    if (size)
+      space->recv_size= size;
+    if (flags != FSP_FLAGS_FCRC32_MASK_MARKER)
+      space->flags= flags;
+  }
   mysql_mutex_unlock(&fil_system.mutex);
 }
 
@@ -1168,16 +1161,12 @@ bool fil_space_t::open(bool create_new_db)
   bool success= true;
   bool skip_read= create_new_db;
 
-  const page_t *page= skip_read
-    ? nullptr
-    : recv_sys.dblwr.find_page(page_id_t{id, 0}, LSN_MAX);
-
   mysql_mutex_lock(&fil_system.mutex);
 
   for (fil_node_t *node= UT_LIST_GET_FIRST(chain); node;
        node= UT_LIST_GET_NEXT(chain, node))
   {
-    if (!node->is_open() && !fil_node_open_file_low(node, page, page))
+    if (!node->is_open() && !fil_node_open_file_low(node))
     {
 err_exit:
       success= false;
@@ -1186,7 +1175,7 @@ err_exit:
 
     if (create_new_db)
     {
-      node->find_metadata();
+      node->find_metadata(node->handle);
       continue;
     }
     if (skip_read)
@@ -1195,7 +1184,7 @@ err_exit:
       continue;
     }
 
-    if (!node->read_page0(page, true))
+    if (!node->read_page0())
     {
       fil_system.n_open--;
       os_file_close(node->handle);
@@ -1204,7 +1193,6 @@ err_exit:
     }
 
     skip_read= true;
-    page= nullptr;
   }
 
   if (!create_new_db)
@@ -1243,9 +1231,6 @@ void fil_system_t::create(ulint hash_size)
 	ut_ad(!is_initialised());
 	ut_ad(!(srv_page_size % FSP_EXTENT_SIZE));
 	ut_ad(srv_page_size);
-	ut_ad(!spaces.array);
-
-	m_initialised = true;
 
 	compile_time_assert(!(UNIV_PAGE_SIZE_MAX % FSP_EXTENT_SIZE_MAX));
 	compile_time_assert(!(UNIV_PAGE_SIZE_MIN % FSP_EXTENT_SIZE_MIN));
@@ -1256,6 +1241,8 @@ void fil_system_t::create(ulint hash_size)
 
 	spaces.create(hash_size);
 
+	need_unflushed_spaces = !write_through && buf_dblwr.need_fsync();
+
 	fil_space_crypt_init();
 #ifdef __linux__
 	ssd.clear();
@@ -1329,13 +1316,12 @@ void fil_system_t::close()
 
   if (is_initialised())
   {
-    m_initialised= false;
     spaces.free();
     mysql_mutex_destroy(&mutex);
     fil_space_crypt_cleanup();
   }
 
-  ut_ad(!spaces.array);
+  ut_ad(!is_initialised());
 
 #ifdef __linux__
   ssd.clear();
@@ -1378,6 +1364,123 @@ ATTRIBUTE_COLD void fil_system_t::extend_to_recv_size()
   mysql_mutex_unlock(&mutex);
 }
 
+ATTRIBUTE_COLD void fil_space_t::reopen_all()
+{
+  mysql_mutex_assert_owner(&fil_system.mutex);
+  fil_system.freeze_space_list++;
+
+  for (fil_space_t &space : fil_system.space_list)
+  {
+    for (fil_node_t *node= UT_LIST_GET_FIRST(space.chain); node;
+         node= UT_LIST_GET_NEXT(chain, node))
+      if (node->is_open())
+        goto need_to_close;
+    continue;
+
+  need_to_close:
+    uint32_t p= space.n_pending.fetch_or(CLOSING, std::memory_order_acquire);
+    if (p & (STOPPING | CLOSING))
+      continue;
+
+    for (fil_node_t *node= UT_LIST_GET_FIRST(space.chain); node;
+         node= UT_LIST_GET_NEXT(chain, node))
+    {
+      if (!node->is_open())
+        continue;
+
+      ulint type= OS_DATA_FILE;
+
+#if defined _WIN32 || defined O_DIRECT
+      switch (FSP_FLAGS_GET_ZIP_SSIZE(space.flags)) {
+      case 1: case 2:
+        type= OS_DATA_FILE_NO_O_DIRECT;
+      }
+#endif
+
+      for (ulint count= 10000; count--;)
+      {
+        p= space.pending();
+
+        if (!(p & CLOSING) || (p & STOPPING))
+          break;
+
+        if (!(p & PENDING) && !node->being_extended)
+        {
+          space.reacquire();
+          mysql_mutex_unlock(&fil_system.mutex);
+          /* Unconditionally flush the file, because
+          fil_system.write_through was updated prematurely,
+          potentially causing some flushes to be lost. */
+          os_file_flush(node->handle);
+          mysql_mutex_lock(&fil_system.mutex);
+          p= space.n_pending.fetch_sub(1, std::memory_order_relaxed) - 1;
+
+          if (!(p & CLOSING) || (p & STOPPING))
+            break;
+
+          if (!(p & PENDING) && !node->being_extended)
+          {
+            ut_a(os_file_close(node->handle));
+            bool success;
+            node->handle= os_file_create(innodb_data_file_key, node->name,
+                                         node->is_raw_disk
+                                         ? OS_FILE_OPEN_RAW : OS_FILE_OPEN,
+                                         OS_FILE_AIO, type,
+                                         srv_read_only_mode, &success);
+            ut_a(success);
+            goto next_file;
+          }
+        }
+
+        space.reacquire();
+        mysql_mutex_unlock(&fil_system.mutex);
+        std::this_thread::sleep_for(std::chrono::microseconds(100));
+        mysql_mutex_lock(&fil_system.mutex);
+        space.release();
+
+        if (!node->is_open())
+          goto next_file;
+      }
+
+      if (!(p & CLOSING) || (p & STOPPING))
+      next_file:
+        continue;
+
+      sql_print_error("InnoDB: Failed to reopen file '%s' due to " UINT32PF
+                      " operations", node->name, p & PENDING);
+    }
+  }
+
+  fil_system.freeze_space_list--;
+}
+
+void fil_system_t::set_write_through(bool write_through)
+{
+  mysql_mutex_lock(&mutex);
+
+  if (write_through != is_write_through())
+  {
+    this->write_through= write_through;
+    fil_space_t::reopen_all();
+    need_unflushed_spaces = !write_through && buf_dblwr.need_fsync();
+  }
+
+  mysql_mutex_unlock(&mutex);
+}
+
+void fil_system_t::set_buffered(bool buffered)
+{
+  mysql_mutex_lock(&mutex);
+
+  if (buffered != is_buffered())
+  {
+    this->buffered= buffered;
+    fil_space_t::reopen_all();
+  }
+
+  mysql_mutex_unlock(&mutex);
+}
+
 /** Close all tablespace files at shutdown */
 void fil_space_t::close_all()
 {
@@ -1398,12 +1501,9 @@ void fil_space_t::close_all()
     for (fil_node_t *node= UT_LIST_GET_FIRST(space.chain); node != NULL;
          node= UT_LIST_GET_NEXT(chain, node))
     {
-
       if (!node->is_open())
-      {
       next:
         continue;
-      }
 
       for (ulint count= 10000; count--;)
       {
@@ -1422,8 +1522,8 @@ void fil_space_t::close_all()
           goto next;
       }
 
-      ib::error() << "File '" << node->name << "' has " << space.referenced()
-                  << " operations";
+      sql_print_error("InnoDB: File '%s' has " UINT32PF " operations",
+                      node->name, space.referenced());
     }
 
     fil_system.detach(&space);
@@ -1496,18 +1596,17 @@ inline void mtr_t::log_file_op(mfile_type_t type, uint32_t space_id,
   m_last= nullptr;
 
   const size_t len= strlen(path);
-  const size_t new_len= new_path ? 1 + strlen(new_path) : 0;
+  const size_t new_len= type == FILE_RENAME ? 1 + strlen(new_path) : 0;
   ut_ad(len > 0);
   byte *const log_ptr= m_log.open(1 + 3/*length*/ + 5/*space_id*/ +
                                   1/*page_no=0*/);
-  *log_ptr= type;
   byte *end= log_ptr + 1;
   end= mlog_encode_varint(end, space_id);
   *end++= 0;
-  const byte *const final_end= end + len + new_len;
-  if (UNIV_LIKELY(final_end >= &log_ptr[16]))
+  if (UNIV_LIKELY(end + len + new_len >= &log_ptr[16]))
   {
-    size_t total_len= final_end - log_ptr - 15;
+    *log_ptr= type;
+    size_t total_len= len + new_len + end - log_ptr - 15;
     if (total_len >= MIN_3BYTE)
       total_len+= 2;
     else if (total_len >= MIN_2BYTE)
@@ -1518,13 +1617,13 @@ inline void mtr_t::log_file_op(mfile_type_t type, uint32_t space_id,
   }
   else
   {
-    *log_ptr= static_cast<byte>(*log_ptr | (final_end - &log_ptr[1]));
+    *log_ptr= static_cast<byte>(type | (end + len + new_len - &log_ptr[1]));
     ut_ad(*log_ptr & 15);
   }
 
   m_log.close(end);
 
-  if (new_path)
+  if (type == FILE_RENAME)
   {
     ut_ad(strchr(new_path, '/'));
     m_log.push(reinterpret_cast<const byte*>(path), uint32_t(len + 1));
@@ -1958,7 +2057,7 @@ fil_ibd_create(
 	file = os_file_create(
 		innodb_data_file_key, path,
 		OS_FILE_CREATE,
-		type, srv_read_only_mode, &success);
+		OS_FILE_AIO, type, srv_read_only_mode, &success);
 
 	if (!success) {
 		/* The following call will print an error message */
@@ -2038,7 +2137,7 @@ err_exit:
 						     FIL_TYPE_TABLESPACE,
 						     crypt_data, mode, true)) {
 		fil_node_t* node = space->add(path, file, size, false, true);
-		node->find_metadata(IF_WIN(,true));
+		IF_WIN(node->find_metadata(), node->find_metadata(file, true));
 		mysql_mutex_unlock(&fil_system.mutex);
 		mtr.start();
 		mtr.set_named_space(space);
@@ -2733,19 +2832,18 @@ static void fil_invalid_page_access_msg(const char *name,
 }
 
 /** Update the data structures on write completion */
-inline void fil_node_t::complete_write()
+void fil_space_t::complete_write()
 {
   mysql_mutex_assert_not_owner(&fil_system.mutex);
 
-  if (space->purpose != FIL_TYPE_TEMPORARY &&
-      srv_file_flush_method != SRV_O_DIRECT_NO_FSYNC &&
-      space->set_needs_flush())
+  if (purpose != FIL_TYPE_TEMPORARY &&
+      fil_system.use_unflushed_spaces() && set_needs_flush())
   {
     mysql_mutex_lock(&fil_system.mutex);
-    if (!space->is_in_unflushed_spaces)
+    if (!is_in_unflushed_spaces)
     {
-      space->is_in_unflushed_spaces= true;
-      fil_system.unflushed_spaces.push_front(*space);
+      is_in_unflushed_spaces= true;
+      fil_system.unflushed_spaces.push_front(*this);
     }
     mysql_mutex_unlock(&fil_system.mutex);
   }
@@ -2787,12 +2885,12 @@ fil_io_t fil_space_t::io(const IORequest &type, os_offset_t offset, size_t len,
 	}
 
 	DBUG_EXECUTE_IF("intermittent_recovery_failure",
-			if (type.is_read() && !(~my_timer_cycles() & 0x3ff0))
+			if (type.is_read() && !(~get_rnd_value() & 0x3ff0))
 			goto io_error;);
 
 	DBUG_EXECUTE_IF("intermittent_read_failure",
 			if (srv_was_started && type.is_read() &&
-			    !(~my_timer_cycles() & 0x3ff0)) goto io_error;);
+			    !(~get_rnd_value() & 0x3ff0)) goto io_error;);
 
 	if (UNIV_LIKELY_NULL(UT_LIST_GET_NEXT(chain, node))) {
 		ut_ad(this == fil_system.sys_space
@@ -2845,7 +2943,7 @@ io_error:
 	if (!type.is_async()) {
 		if (type.is_write()) {
 release_sync_write:
-			node->complete_write();
+			complete_write();
 release:
 			release();
 			goto func_exit;
@@ -2865,21 +2963,28 @@ void IORequest::write_complete(int io_error) const
 {
   ut_ad(fil_validate_skip());
   ut_ad(node);
+  fil_space_t *space= node->space;
   ut_ad(is_write());
-  node->complete_write();
 
   if (!bpage)
   {
     ut_ad(!srv_read_only_mode);
     if (type == IORequest::DBLWR_BATCH)
+    {
       buf_dblwr.flush_buffered_writes_completed(*this);
+      /* Above, we already invoked os_file_flush() on the
+      doublewrite buffer if needed. */
+      goto func_exit;
+    }
     else
       ut_ad(type == IORequest::WRITE_ASYNC);
   }
   else
     buf_page_write_complete(*this, io_error);
 
-  node->space->release();
+  space->complete_write();
+ func_exit:
+  space->release();
 }
 
 void IORequest::read_complete(int io_error) const
@@ -2889,17 +2994,12 @@ void IORequest::read_complete(int io_error) const
   ut_ad(is_read());
   ut_ad(bpage);
 
-  /* IMPORTANT: since i/o handling for reads will read also the insert
-  buffer in fil_system.sys_space, we have to be very careful not to
-  introduce deadlocks. We never close fil_system.sys_space data files
-  and never issue asynchronous reads of change buffer pages. */
   const page_id_t id(bpage->id());
 
   if (UNIV_UNLIKELY(io_error != 0))
   {
     sql_print_error("InnoDB: Read error %d of page " UINT32PF " in file %s",
                     io_error, id.page_no(), node->name);
-    recv_sys.free_corrupted_page(id, *node);
     buf_pool.corrupted_evict(bpage, buf_page_t::READ_FIX);
   corrupted:
     if (recv_recovery_is_on() && !srv_force_recovery)
@@ -2909,8 +3009,13 @@ void IORequest::read_complete(int io_error) const
       mysql_mutex_unlock(&recv_sys.mutex);
     }
   }
-  else if (bpage->read_complete(*node))
+  else if (dberr_t err= bpage->read_complete(*node))
+  {
+    if (err != DB_FAIL)
+      ib::error() << "Failed to read page " << id.page_no()
+                  << " from file '" << node->name << "': " << err;
     goto corrupted;
+  }
 
   node->space->release();
 }
@@ -2919,14 +3024,6 @@ void IORequest::read_complete(int io_error) const
 possibly cached by the OS. */
 void fil_flush_file_spaces()
 {
-  if (srv_file_flush_method == SRV_O_DIRECT_NO_FSYNC)
-  {
-    ut_d(mysql_mutex_lock(&fil_system.mutex));
-    ut_ad(fil_system.unflushed_spaces.empty());
-    ut_d(mysql_mutex_unlock(&fil_system.mutex));
-    return;
-  }
-
 rescan:
   mysql_mutex_lock(&fil_system.mutex);
 
