@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2013, 2022, MariaDB Corporation.
+Copyright (c) 2013, 2023, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -48,10 +48,6 @@ Created 11/5/1995 Heikki Tuuri
 #define	BUF_GET_IF_IN_POOL	11	/*!< get if in pool */
 #define BUF_PEEK_IF_IN_POOL	12	/*!< get if in pool, do not make
 					the block young in the LRU list */
-#define BUF_GET_IF_IN_POOL_OR_WATCH	15
-					/*!< Get the page only if it's in the
-					buffer pool, if not then set a watch
-					on the page. */
 #define BUF_GET_POSSIBLY_FREED		16
 					/*!< Like BUF_GET, but do not mind
 					if the file page has been freed. */
@@ -194,62 +190,34 @@ be implemented at a higher level.  In other words, all possible
 accesses to a given page through this function must be protected by
 the same set of mutexes or latches.
 @param page_id   page identifier
+@param zip_size  ROW_FORMAT=COMPRESSED page size in bytes
 @return pointer to the block, s-latched */
-buf_page_t *buf_page_get_zip(const page_id_t page_id);
+buf_page_t *buf_page_get_zip(const page_id_t page_id, ulint zip_size);
 
 /** Get access to a database page. Buffered redo log may be applied.
 @param[in]	page_id			page id
 @param[in]	zip_size		ROW_FORMAT=COMPRESSED page size, or 0
-@param[in]	rw_latch		latch mode
+@param[in]	rw_latch		RW_S_LATCH, RW_X_LATCH, RW_NO_LATCH
 @param[in]	guess			guessed block or NULL
 @param[in]	mode			BUF_GET, BUF_GET_IF_IN_POOL,
-BUF_PEEK_IF_IN_POOL, or BUF_GET_IF_IN_POOL_OR_WATCH
+or BUF_PEEK_IF_IN_POOL
 @param[in,out]	mtr			mini-transaction
 @param[out]	err			DB_SUCCESS or error code
-@param[in]	allow_ibuf_merge	Allow change buffer merge to happen
-@return pointer to the block
-@retval nullptr	if the block is corrupted or unavailable */
+@return pointer to the block or NULL */
 buf_block_t*
 buf_page_get_gen(
 	const page_id_t		page_id,
 	ulint			zip_size,
-	rw_lock_type_t		rw_latch,
+	ulint			rw_latch,
 	buf_block_t*		guess,
 	ulint			mode,
 	mtr_t*			mtr,
-	dberr_t*		err = nullptr,
-	bool			allow_ibuf_merge = false)
-	MY_ATTRIBUTE((nonnull(6), warn_unused_result));
-
-/** This is the low level function used to get access to a database page.
-@param[in]	page_id			page id
-@param[in]	zip_size		ROW_FORMAT=COMPRESSED page size, or 0
-@param[in]	rw_latch		latch mode
-@param[in]	guess			guessed block or NULL
-@param[in]	mode			BUF_GET, BUF_GET_IF_IN_POOL,
-BUF_PEEK_IF_IN_POOL, or BUF_GET_IF_IN_POOL_OR_WATCH
-@param[in,out]	mtr			mini-transaction, or NULL if a
-					block with page_id is to be evicted
-@param[out]	err			DB_SUCCESS or error code
-@param[in]	allow_ibuf_merge	Allow change buffer merge to happen
-@return pointer to the block
-@retval nullptr	if the block is corrupted or unavailable */
-buf_block_t*
-buf_page_get_low(
-	const page_id_t		page_id,
-	ulint			zip_size,
-	rw_lock_type_t		rw_latch,
-	buf_block_t*		guess,
-	ulint			mode,
-	mtr_t*			mtr,
-	dberr_t*		err,
-	bool			allow_ibuf_merge)
-	MY_ATTRIBUTE((nonnull(6), warn_unused_result));
+	dberr_t*		err = nullptr);
 
 /** Initialize a page in the buffer pool. The page is usually not read
 from a file even if it cannot be found in the buffer buf_pool. This is one
 of the functions which perform to a block a state transition NOT_USED => LRU
-(the other is buf_page_get_low()).
+(the other is buf_page_get_gen()).
 @param[in,out]	space		space object
 @param[in]	offset		offset of the tablespace
 @param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
@@ -308,21 +276,13 @@ buf_block_modify_clock_inc(
 @return whether the buffer is all zeroes */
 bool buf_is_zeroes(st_::span<const byte> buf);
 
-/** Reason why buf_page_is_corrupted() fails */
-enum buf_page_is_corrupted_reason
-{
-  CORRUPTED_FUTURE_LSN= -1,
-  NOT_CORRUPTED= 0,
-  CORRUPTED_OTHER
-};
-
 /** Check if a page is corrupt.
 @param check_lsn   whether FIL_PAGE_LSN should be checked
 @param read_buf    database page
 @param fsp_flags   contents of FIL_SPACE_FLAGS
 @return whether the page is corrupted */
-buf_page_is_corrupted_reason
-buf_page_is_corrupted(bool check_lsn, const byte *read_buf, uint32_t fsp_flags)
+bool buf_page_is_corrupted(bool check_lsn, const byte *read_buf,
+                           uint32_t fsp_flags)
   MY_ATTRIBUTE((warn_unused_result));
 
 /** Read the key version from the page. In full crc32 format,
@@ -392,8 +352,8 @@ void buf_page_print(const byte* read_buf, ulint zip_size = 0)
 	ATTRIBUTE_COLD __attribute__((nonnull));
 /********************************************************************//**
 Decompress a block.
-@return true if successful */
-bool
+@return TRUE if successful */
+ibool
 buf_zip_decompress(
 /*===============*/
 	buf_block_t*	block,	/*!< in/out: block */
@@ -535,18 +495,16 @@ public:
   static constexpr uint32_t REMOVE_HASH= 2;
   /** smallest state() of a buffer page that is freed in the tablespace */
   static constexpr uint32_t FREED= 3;
+  /* unused state: 1U<<29 */
   /** smallest state() for a block that belongs to buf_pool.LRU */
-  static constexpr uint32_t UNFIXED= 1U << 29;
-  /** smallest state() of a block for which buffered changes may exist */
-  static constexpr uint32_t IBUF_EXIST= 2U << 29;
+  static constexpr uint32_t UNFIXED= 2U << 29;
   /** smallest state() of a (re)initialized page (no doublewrite needed) */
   static constexpr uint32_t REINIT= 3U << 29;
   /** smallest state() for an io-fixed block */
   static constexpr uint32_t READ_FIX= 4U << 29;
+  /* unused state: 5U<<29 */
   /** smallest state() for a write-fixed block */
-  static constexpr uint32_t WRITE_FIX= 5U << 29;
-  /** smallest state() for a write-fixed block with buffered changes */
-  static constexpr uint32_t WRITE_FIX_IBUF= 6U << 29;
+  static constexpr uint32_t WRITE_FIX= 6U << 29;
   /** smallest state() for a write-fixed block (no doublewrite was used) */
   static constexpr uint32_t WRITE_FIX_REINIT= 7U << 29;
   /** buf_pool.LRU status mask in state() */
@@ -558,8 +516,7 @@ public:
   byte *frame;
   /* @} */
   /** ROW_FORMAT=COMPRESSED page; zip.data (but not the data it points to)
-  is also protected by buf_pool.mutex;
-  !frame && !zip.data means an active buf_pool.watch */
+  is also protected by buf_pool.mutex */
   page_zip_des_t zip;
 #ifdef UNIV_DEBUG
   /** whether this->list is in buf_pool.zip_hash; protected by buf_pool.mutex */
@@ -640,10 +597,10 @@ public:
   void init(uint32_t state, page_id_t id)
   {
     ut_ad(state < REMOVE_HASH || state >= UNFIXED);
-    ut_ad(!lock.is_locked_or_waiting());
     id_= id;
     zip.fix= state;
     oldest_modification_= 0;
+    lock.init();
     ut_d(in_zip_hash= false);
     ut_d(in_free_list= false);
     ut_d(in_LRU_list= false);
@@ -665,49 +622,30 @@ public:
 public:
   const page_id_t &id() const { return id_; }
   uint32_t state() const { return zip.fix; }
-  static uint32_t buf_fix_count(uint32_t s)
-  { ut_ad(s >= FREED); return s < UNFIXED ? (s - FREED) : (~LRU_MASK & s); }
-
-  uint32_t buf_fix_count() const { return buf_fix_count(state()); }
-  /** Check if a file block is io-fixed.
-  @param s   state()
-  @return whether s corresponds to an io-fixed block */
-  static bool is_io_fixed(uint32_t s)
-  { ut_ad(s >= FREED); return s >= READ_FIX; }
-  /** Check if a file block is read-fixed.
-  @param s   state()
-  @return whether s corresponds to a read-fixed block */
-  static bool is_read_fixed(uint32_t s)
-  { return is_io_fixed(s) && s < WRITE_FIX; }
-  /** Check if a file block is write-fixed.
-  @param s   state()
-  @return whether s corresponds to a write-fixed block */
-  static bool is_write_fixed(uint32_t s)
-  { ut_ad(s >= FREED); return s >= WRITE_FIX; }
-
+  uint32_t buf_fix_count() const
+  {
+    uint32_t f= state();
+    ut_ad(f >= FREED);
+    return f < UNFIXED ? (f - FREED) : (~LRU_MASK & f);
+  }
   /** @return whether this block is read or write fixed;
   read_complete() or write_complete() will always release
   the io-fix before releasing U-lock or X-lock */
-  bool is_io_fixed() const { return is_io_fixed(state()); }
+  bool is_io_fixed() const
+  { const auto s= state(); ut_ad(s >= FREED); return s >= READ_FIX; }
   /** @return whether this block is write fixed;
   write_complete() will always release the write-fix before releasing U-lock */
-  bool is_write_fixed() const { return is_write_fixed(state()); }
-  /** @return whether this block is read fixed */
-  bool is_read_fixed() const { return is_read_fixed(state()); }
+  bool is_write_fixed() const { return state() >= WRITE_FIX; }
+  /** @return whether this block is read fixed; this should never hold
+  when a thread is holding the block lock in any mode */
+  bool is_read_fixed() const { return is_io_fixed() && !is_write_fixed(); }
 
   /** @return if this belongs to buf_pool.unzip_LRU */
   bool belongs_to_unzip_LRU() const
   { return UNIV_LIKELY_NULL(zip.data) && frame; }
 
-  static bool is_freed(uint32_t s) { ut_ad(s >= FREED); return s < UNFIXED; }
-  bool is_freed() const { return is_freed(state()); }
-  static bool is_ibuf_exist(uint32_t s)
-  {
-    ut_ad(s >= UNFIXED);
-    ut_ad(s < READ_FIX);
-    return (s & LRU_MASK) == IBUF_EXIST;
-  }
-  bool is_ibuf_exist() const { return is_ibuf_exist(state()); }
+  bool is_freed() const
+  { const auto s= state(); ut_ad(s >= FREED); return s < UNFIXED; }
   bool is_reinit() const { return !(~state() & REINIT); }
 
   void set_reinit(uint32_t prev_state)
@@ -718,29 +656,10 @@ public:
     ut_ad(s < prev_state + UNFIXED);
   }
 
-  void set_ibuf_exist()
-  {
-    ut_ad(lock.is_write_locked());
-    ut_ad(id() < page_id_t(SRV_SPACE_ID_UPPER_BOUND, 0));
-    const auto s= state();
-    ut_ad(s >= UNFIXED);
-    ut_ad(s < READ_FIX);
-    ut_ad(s < IBUF_EXIST || s >= REINIT);
-    zip.fix.fetch_add(IBUF_EXIST - (LRU_MASK & s));
-  }
-  void clear_ibuf_exist()
-  {
-    ut_ad(lock.is_write_locked());
-    ut_ad(id() < page_id_t(SRV_SPACE_ID_UPPER_BOUND, 0));
-    ut_d(const auto s=) zip.fix.fetch_sub(IBUF_EXIST - UNFIXED);
-    ut_ad(s >= IBUF_EXIST);
-    ut_ad(s < REINIT);
-  }
-
   uint32_t read_unfix(uint32_t s)
   {
     ut_ad(lock.is_write_locked());
-    ut_ad(s == UNFIXED + 1 || s == IBUF_EXIST + 1 || s == REINIT + 1);
+    ut_ad(s == UNFIXED + 1 || s == REINIT + 1);
     uint32_t old_state= zip.fix.fetch_add(s - READ_FIX);
     ut_ad(old_state >= READ_FIX);
     ut_ad(old_state < WRITE_FIX);
@@ -786,8 +705,9 @@ public:
   /** Complete a read of a page.
   @param node     data file
   @return whether the operation succeeded
-  @retval DB_PAGE_CORRUPTED    if the checksum or the page ID is incorrect
-  @retval DB_DECRYPTION_FAILED if the page cannot be decrypted */
+  @retval DB_PAGE_CORRUPTED    if the checksum fails
+  @retval DB_DECRYPTION_FAILED if the page cannot be decrypted
+  @retval DB_FAIL              if the page contains the wrong ID */
   dberr_t read_complete(const fil_node_t &node);
 
   /** Release a write fix after a page write was completed.
@@ -829,7 +749,7 @@ public:
   uint32_t fix(uint32_t count= 1)
   {
     ut_ad(count);
-    ut_ad(count < IBUF_EXIST);
+    ut_ad(count < REINIT);
     uint32_t f= zip.fix.fetch_add(count);
     ut_ad(f >= FREED);
     ut_ad(!((f ^ (f + 1)) & LRU_MASK));
@@ -1369,11 +1289,10 @@ public:
   /** Release and evict a corrupted page.
   @param bpage    x-latched page that was found corrupted
   @param state    expected current state of the page */
-  ATTRIBUTE_COLD void corrupted_evict(buf_page_t *bpage, uint32_t state)
-    noexcept;
+  ATTRIBUTE_COLD void corrupted_evict(buf_page_t *bpage, uint32_t state);
 
   /** Release a memory block to the buffer pool. */
-  ATTRIBUTE_COLD void free_block(buf_block_t *block) noexcept;
+  ATTRIBUTE_COLD void free_block(buf_block_t *block);
 
 #ifdef UNIV_DEBUG
   /** Find a block that points to a ROW_FORMAT=COMPRESSED page
@@ -1434,117 +1353,11 @@ public:
   }
 
 public:
-  /** page_fix() mode of operation */
-  enum page_fix_conflicts{
-    /** Fetch if in the buffer pool, also blocks marked as free */
-    FIX_ALSO_FREED= -1,
-    /** Fetch, waiting for page read completion */
-    FIX_WAIT_READ,
-    /** Fetch, but avoid any waits for */
-    FIX_NOWAIT
-  };
-
-  /** Look up and buffer-fix a page.
-  Note: If the page is read-fixed (being read into the buffer pool),
-  we would have to wait for the page latch before determining if the page
-  is accessible (it could be corrupted and have been evicted again).
-  If the caller is holding other page latches so that waiting for this
-  page latch could lead to lock order inversion (latching order violation),
-  the mode c=FIX_WAIT_READ must not be used.
-  @param id        page identifier
-  @param err       error code (will only be assigned when returning nullptr)
-  @param c         how to handle conflicts
-  @return undo log page, buffer-fixed
-  @retval -1       if c=FIX_NOWAIT and buffer-fixing would require waiting
-  @retval nullptr  if the undo page was corrupted or freed */
-  buf_block_t *page_fix(const page_id_t id, dberr_t *err,
-                        page_fix_conflicts c);
-
-  buf_block_t *page_fix(const page_id_t id)
-  { return page_fix(id, nullptr, FIX_WAIT_READ); }
-
-
-  /** Decompress a page and relocate the block descriptor
-  @param b      buffer-fixed compressed-only ROW_FORMAT=COMPRESSED page
-  @param chain  hash table chain for b->id().fold()
-  @return the decompressed block, x-latched and read-fixed
-  @retval nullptr if the decompression failed (b->unfix() will be invoked) */
-  ATTRIBUTE_COLD __attribute__((nonnull, warn_unused_result))
-  buf_block_t *unzip(buf_page_t *b, hash_chain &chain);
-
   /** @return whether the buffer pool contains a page
-  @tparam allow_watch  whether to allow watch_is_sentinel()
   @param page_id       page identifier
   @param chain         hash table chain for page_id.fold() */
-  template<bool allow_watch= false>
-  TRANSACTIONAL_INLINE
-  bool page_hash_contains(const page_id_t page_id, hash_chain &chain)
-  {
-    transactional_shared_lock_guard<page_hash_latch> g
-      {page_hash.lock_get(chain)};
-    buf_page_t *bpage= page_hash.get(page_id, chain);
-    if (bpage >= &watch[0] && bpage < &watch[UT_ARR_SIZE(watch)])
-    {
-      ut_ad(!bpage->in_zip_hash);
-      ut_ad(!bpage->zip.data);
-      if (!allow_watch)
-        bpage= nullptr;
-    }
-    return bpage;
-  }
-
-  /** Determine if a block is a sentinel for a buffer pool watch.
-  @param bpage page descriptor
-  @return whether bpage a sentinel for a buffer pool watch */
-  bool watch_is_sentinel(const buf_page_t &bpage)
-  {
-#ifdef SAFE_MUTEX
-    DBUG_ASSERT(mysql_mutex_is_owner(&mutex) ||
-                page_hash.lock_get(page_hash.cell_get(bpage.id().fold())).
-                is_locked());
-#endif /* SAFE_MUTEX */
-    ut_ad(bpage.in_file());
-    if (&bpage < &watch[0] || &bpage >= &watch[array_elements(watch)])
-      return false;
-    ut_ad(!bpage.in_zip_hash);
-    ut_ad(!bpage.zip.data);
-    return true;
-  }
-
-  /** Check if a watched page has been read.
-  This may only be called after !watch_set() and before invoking watch_unset().
-  @param id   page identifier
-  @return whether the page was read to the buffer pool */
-  TRANSACTIONAL_INLINE
-  bool watch_occurred(const page_id_t id)
-  {
-    hash_chain &chain= page_hash.cell_get(id.fold());
-    transactional_shared_lock_guard<page_hash_latch> g
-      {page_hash.lock_get(chain)};
-    /* The page must exist because watch_set() increments buf_fix_count. */
-    return !watch_is_sentinel(*page_hash.get(id, chain));
-  }
-
-  /** Register a watch for a page identifier.
-  @param id         page identifier
-  @param chain      page_hash.cell_get(id.fold())
-  @return a buffer page corresponding to id
-  @retval nullptr   if the block was not present in page_hash */
-  buf_page_t *watch_set(const page_id_t id, hash_chain &chain);
-
-  /** Stop watching whether a page has been read in.
-  watch_set(id) must have returned nullptr before.
-  @param id         page identifier
-  @param chain      unlocked hash table chain */
-  void watch_unset(const page_id_t id, hash_chain &chain);
-
-  /** Remove the sentinel block for the watch before replacing it with a
-  real block. watch_unset() or watch_occurred() will notice
-  that the block has been replaced with the real block.
-  @param w          sentinel
-  @param chain      locked hash table chain
-  @return           w->state() */
-  inline uint32_t watch_remove(buf_page_t *w, hash_chain &chain);
+  TRANSACTIONAL_TARGET
+  bool page_hash_contains(const page_id_t page_id, hash_chain &chain);
 
   /** @return whether less than 1/4 of the buffer pool is available */
   TPOOL_SUPPRESS_TSAN
@@ -1552,7 +1365,7 @@ public:
   {
     return !recv_recovery_is_on() &&
       UT_LIST_GET_LEN(free) + UT_LIST_GET_LEN(LRU) <
-        (n_chunks_new * chunks->size) / 4;
+        n_chunks_new / 4 * chunks->size;
   }
 
   /** @return whether the buffer pool is running low */
@@ -1593,6 +1406,8 @@ public:
   /** Number of pages to read ahead */
   static constexpr uint32_t READ_AHEAD_PAGES= 64;
 
+  /** Buffer pool mutex */
+  alignas(CPU_LEVEL1_DCACHE_LINESIZE) mysql_mutex_t mutex;
   /** current statistics; protected by mutex */
   buf_pool_stat_t stat;
   /** old statistics; protected by mutex */
@@ -1739,16 +1554,6 @@ public:
     inline void write_unlock_all();
   };
 
-  /** Buffer pool mutex */
-  alignas(CPU_LEVEL1_DCACHE_LINESIZE) mysql_mutex_t mutex;
-
-  /** innodb_lru_scan_depth; number of blocks scanned in LRU flush batch;
-  protected by buf_pool_t::mutex */
-  ulong LRU_scan_depth;
-  /** innodb_flush_neighbors; whether or not to flush neighbors of a block;
-  protected by buf_pool_t::mutex */
-  ulong flush_neighbors;
-
   /** Hash table of file pages (buf_page_t::in_file() holds),
   indexed by page_id_t. Protected by both mutex and page_hash.lock_get(). */
   page_hash_table page_hash;
@@ -1756,8 +1561,8 @@ public:
   /** map of block->frame to buf_block_t blocks that belong
   to buf_buddy_alloc(); protected by buf_pool.mutex */
   hash_table_t zip_hash;
-  /** number of pending unzip() */
-  Atomic_counter<ulint> n_pend_unzip;
+	Atomic_counter<ulint>
+			n_pend_unzip;	/*!< number of pending decompressions */
 
 	time_t		last_printout_time;
 					/*!< when buf_print_io was last time
@@ -1947,9 +1752,6 @@ public:
 # error "BUF_BUDDY_LOW > UNIV_ZIP_SIZE_MIN"
 #endif
 
-  /** Sentinels to detect if pages are read into the buffer pool while
-  a delete-buffering operation is pending. Protected by mutex. */
-  buf_page_t watch[innodb_purge_threads_MAX + 1];
   /** Reserve a buffer. */
   buf_tmp_buffer_t *io_buf_reserve(bool wait_for_reads)
   { return io_buf.reserve(wait_for_reads); }
