@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2013, 2022, MariaDB Corporation.
+Copyright (c) 2013, 2023, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -48,38 +48,6 @@ struct named_spaces_tag_t;
 
 using space_list_t= ilist<fil_space_t, space_list_tag_t>;
 
-// Forward declaration
-extern my_bool srv_use_doublewrite_buf;
-
-/** Possible values of innodb_flush_method */
-enum srv_flush_t
-{
-  /** fsync, the default */
-  SRV_FSYNC= 0,
-  /** open log files in O_DSYNC mode */
-  SRV_O_DSYNC,
-  /** do not call os_file_flush() when writing data files, but do flush
-  after writing to log files */
-  SRV_LITTLESYNC,
-  /** do not flush after writing */
-  SRV_NOSYNC,
-  /** Open or create files with O_DIRECT. This implies using
-  unbuffered I/O but still fdatasync(), because some filesystems might
-  not flush meta-data on write completion */
-  SRV_O_DIRECT,
-  /** Like O_DIRECT, but skip fdatasync(), assuming that the data is
-  durable on write completion */
-  SRV_O_DIRECT_NO_FSYNC
-#ifdef _WIN32
-  /** Traditional Windows appoach to open all files without caching,
-  and do FileFlushBuffers() */
-  ,SRV_ALL_O_DIRECT_FSYNC
-#endif
-};
-
-/** innodb_flush_method */
-extern ulong srv_file_flush_method;
-
 /** Undo tablespaces starts with space_id. */
 extern uint32_t srv_undo_space_id_start;
 /** The number of UNDO tablespaces that are open and ready to use. */
@@ -513,9 +481,8 @@ public:
   written while the space ID is being updated in each page. */
   inline void set_imported();
 
-  /** Report the tablespace as corrupted
-  @return whether this was the first call */
-  ATTRIBUTE_COLD bool set_corrupted() const noexcept;
+  /** Report the tablespace as corrupted */
+  ATTRIBUTE_COLD void set_corrupted() const;
 
   /** @return whether the storage device is rotational (HDD, not SSD) */
   inline bool is_rotational() const;
@@ -527,6 +494,9 @@ public:
   /** Close each file. Only invoked on fil_system.temp_space. */
   void close();
 
+  /** Note that operations on the tablespace must stop. */
+  inline void set_stopping();
+
   /** Drop the tablespace and wait for any pending operations to cease
   @param id               tablespace identifier
   @param detached_handle  pointer to file to be closed later, or nullptr
@@ -585,14 +555,32 @@ public:
   /** Clear the NEEDS_FSYNC flag */
   void clear_flush()
   {
+#if defined __GNUC__ && (defined __i386__ || defined __x86_64__)
+    static_assert(NEEDS_FSYNC == 1U << 28, "compatibility");
+    __asm__ __volatile__("lock btrl $28, %0" : "+m" (n_pending));
+#elif defined _MSC_VER && (defined _M_IX86 || defined _M_X64)
+    static_assert(NEEDS_FSYNC == 1U << 28, "compatibility");
+    _interlockedbittestandreset(reinterpret_cast<volatile long*>
+                                (&n_pending), 28);
+#else
     n_pending.fetch_and(~NEEDS_FSYNC, std::memory_order_release);
+#endif
   }
 
 private:
   /** Clear the CLOSING flag */
   void clear_closing()
   {
+#if defined __GNUC__ && (defined __i386__ || defined __x86_64__)
+    static_assert(CLOSING == 1U << 29, "compatibility");
+    __asm__ __volatile__("lock btrl $29, %0" : "+m" (n_pending));
+#elif defined _MSC_VER && (defined _M_IX86 || defined _M_X64)
+    static_assert(CLOSING == 1U << 29, "compatibility");
+    _interlockedbittestandreset(reinterpret_cast<volatile long*>
+                                (&n_pending), 29);
+#else
     n_pending.fetch_and(~CLOSING, std::memory_order_relaxed);
+#endif
   }
 
   /** @return pending operations (and flags) */
@@ -625,11 +613,12 @@ private:
   }
 
 public:
+  /** Reopen all files on set_write_through() or set_buffered(). */
+  static void reopen_all();
   /** Try to close a file to adhere to the innodb_open_files limit.
-  @param ignore_space Ignore the tablespace which is acquired by caller
   @param print_info   whether to diagnose why a file cannot be closed
   @return whether a file was closed */
-  static bool try_to_close(fil_space_t *ignore_space, bool print_info);
+  static bool try_to_close(bool print_info);
 
   /** Close all tablespace files at shutdown */
   static void close_all();
@@ -970,10 +959,8 @@ public:
   void flush_low();
 
   /** Read the first page of a data file.
-  @param dpage   copy of a first page, from the doublewrite buffer, or nullptr
-  @param no_lsn  whether to skip the FIL_PAGE_LSN check
   @return whether the page was found valid */
-  bool read_page0(const byte *dpage, bool no_lsn) noexcept;
+  bool read_page0();
 
   /** Determine the next tablespace for encryption key rotation.
   @param space    current tablespace (nullptr to start from the beginning)
@@ -1018,6 +1005,9 @@ public:
   /** @return the tablespace name (databasename/tablename) */
   name_type name() const;
 
+  /** Update the data structures on write completion */
+  void complete_write();
+
 private:
   /** @return whether the file is usable for io() */
   ATTRIBUTE_COLD bool prepare_acquired();
@@ -1070,12 +1060,16 @@ struct fil_node_t final
   bool is_open() const { return handle != OS_FILE_CLOSED; }
 
   /** Read the first page of a data file.
-  @param dpage   copy of a first page, from the doublewrite buffer, or nullptr
-  @param no_lsn  whether to skip the FIL_PAGE_LSN check
   @return whether the page was found valid */
-  bool read_page0(const byte *dpage, bool no_lsn) noexcept;
-  /** Determine some file metadata when creating or reading the file. */
-  void find_metadata(IF_WIN(,bool create= false)) noexcept;
+  bool read_page0();
+
+  /** Determine some file metadata when creating or reading the file.
+  @param file   the file that is being created, or OS_FILE_CLOSED */
+  void find_metadata(os_file_t file= OS_FILE_CLOSED
+#ifndef _WIN32
+                     , bool create= false, struct stat *statbuf= nullptr
+#endif
+                     );
 
   /** Close the file handle. */
   void close();
@@ -1086,9 +1080,6 @@ struct fil_node_t final
   @return detached handle or OS_FILE_CLOSED */
   inline pfs_os_file_t close_to_free(bool detach_handle= false);
 
-  /** Update the data structures on write completion */
-  inline void complete_write();
-
 private:
   /** Does stuff common for close() and detach() */
   void prepare_to_close_or_detach();
@@ -1096,8 +1087,7 @@ private:
 
 inline bool fil_space_t::use_doublewrite() const
 {
-  return !UT_LIST_GET_FIRST(chain)->atomic_write && srv_use_doublewrite_buf &&
-    buf_dblwr.is_created();
+  return !UT_LIST_GET_FIRST(chain)->atomic_write && buf_dblwr.in_use();
 }
 
 inline void fil_space_t::set_imported()
@@ -1264,11 +1254,11 @@ constexpr uint16_t FIL_PAGE_RTREE= 17854;
 constexpr uint16_t FIL_PAGE_UNDO_LOG= 2;
 /** Index node (of file-in-file metadata) */
 constexpr uint16_t FIL_PAGE_INODE= 3;
-/** Insert buffer free list */
+/** Former change buffer free list */
 constexpr uint16_t FIL_PAGE_IBUF_FREE_LIST= 4;
 /** Freshly allocated page */
 constexpr uint16_t FIL_PAGE_TYPE_ALLOCATED= 0;
-/** Change buffer bitmap (pages n*innodb_page_size+1) */
+/** Former change buffer bitmap pages (pages n*innodb_page_size+1) */
 constexpr uint16_t FIL_PAGE_IBUF_BITMAP= 5;
 /** System page */
 constexpr uint16_t FIL_PAGE_TYPE_SYS= 6;
@@ -1358,9 +1348,9 @@ struct fil_system_t
     Some members may require late initialisation, thus we just mark object as
     uninitialised. Real initialisation happens in create().
   */
-  fil_system_t() : m_initialised(false) {}
+  fil_system_t() {}
 
-  bool is_initialised() const { return m_initialised; }
+  bool is_initialised() const { return spaces.array; }
 
   /**
     Create the file system interface at database start.
@@ -1373,8 +1363,6 @@ struct fil_system_t
   void close();
 
 private:
-  bool m_initialised;
-
   /** Points to the last opened space in space_list. Protected with
   fil_system.mutex. */
   fil_space_t *space_list_last_opened= nullptr;
@@ -1409,6 +1397,33 @@ public:
 	fil_space_t*	temp_space;	/*!< The innodb_temporary tablespace */
   /** Map of fil_space_t::id to fil_space_t* */
   hash_table_t spaces;
+
+  /** false=invoke fsync() or fdatasync() on data files before checkpoint;
+  true=each write is durable (O_DSYNC) */
+  my_bool write_through;
+  /** whether data files are buffered (not O_DIRECT) */
+  my_bool buffered;
+  /** whether fdatasync() is needed on data files */
+  Atomic_relaxed<bool> need_unflushed_spaces;
+
+  /** Try to enable or disable write-through of data files */
+  void set_write_through(bool write_through);
+  /** Update innodb_doublewrite */
+  void set_use_doublewrite(ulong use)
+  {
+    buf_dblwr.set_use(use);
+    need_unflushed_spaces= !write_through && buf_dblwr.need_fsync();
+  }
+
+  /** Try to enable or disable file system caching of data files */
+  void set_buffered(bool buffered);
+
+  TPOOL_SUPPRESS_TSAN bool is_write_through() const { return write_through; }
+  TPOOL_SUPPRESS_TSAN bool is_buffered() const { return buffered; }
+
+  /** @return whether to update unflushed_spaces */
+  bool use_unflushed_spaces() const { return need_unflushed_spaces; }
+
   /** tablespaces for which fil_space_t::needs_flush() holds */
   sized_ilist<fil_space_t, unflushed_spaces_tag_t> unflushed_spaces;
   /** number of currently open files; protected by mutex */
@@ -1513,18 +1528,28 @@ inline void fil_space_t::reacquire()
 #endif /* SAFE_MUTEX */
 }
 
+/** Note that operations on the tablespace must stop. */
+inline void fil_space_t::set_stopping()
+{
+  mysql_mutex_assert_owner(&fil_system.mutex);
+#if defined __GNUC__ && (defined __i386__ || defined __x86_64__)
+  static_assert(STOPPING_WRITES == 1U << 30, "compatibility");
+  __asm__ __volatile__("lock btsl $30, %0" : "+m" (n_pending));
+#elif defined _MSC_VER && (defined _M_IX86 || defined _M_X64)
+  static_assert(STOPPING_WRITES == 1U << 30, "compatibility");
+  _interlockedbittestandset(reinterpret_cast<volatile long*>(&n_pending), 30);
+#else
+  n_pending.fetch_or(STOPPING_WRITES, std::memory_order_relaxed);
+#endif
+}
+
 /** Flush pending writes from the file system cache to the file. */
 template<bool have_reference> inline void fil_space_t::flush()
 {
   mysql_mutex_assert_not_owner(&fil_system.mutex);
   ut_ad(!have_reference || (pending() & PENDING));
   ut_ad(purpose == FIL_TYPE_TABLESPACE || purpose == FIL_TYPE_IMPORT);
-  if (srv_file_flush_method == SRV_O_DIRECT_NO_FSYNC)
-  {
-    ut_ad(!is_in_unflushed_spaces);
-    ut_ad(!needs_flush());
-  }
-  else if (have_reference)
+  if (have_reference)
     flush_low();
   else
   {
@@ -1542,7 +1567,7 @@ inline uint32_t fil_space_t::get_size()
   if (!size)
   {
     mysql_mutex_lock(&fil_system.mutex);
-    read_page0(nullptr, false);
+    read_page0();
     mysql_mutex_unlock(&fil_system.mutex);
   }
   return size;
