@@ -42,7 +42,6 @@ Created 11/26/1995 Heikki Tuuri
 #ifdef HAVE_PMEM
 void (*mtr_t::commit_logger)(mtr_t *, std::pair<lsn_t,page_flush_ahead>);
 #endif
-
 std::pair<lsn_t,mtr_t::page_flush_ahead> (*mtr_t::finisher)(mtr_t *, size_t);
 unsigned mtr_t::spin_wait_delay;
 
@@ -50,7 +49,7 @@ void mtr_t::finisher_update()
 {
   ut_ad(log_sys.latch_have_wr());
 #ifdef HAVE_PMEM
-  if (log_sys.is_mmap())
+  if (log_sys.is_pmem())
   {
     commit_logger= mtr_t::commit_log<true>;
     finisher= spin_wait_delay
@@ -204,7 +203,6 @@ void mtr_t::start()
 
   m_made_dirty= false;
   m_latch_ex= false;
-  m_inside_ibuf= false;
   m_modifications= false;
   m_log_mode= MTR_LOG_ALL;
   ut_d(m_user_space_id= TRX_SYS_SPACE);
@@ -352,11 +350,11 @@ inline lsn_t log_t::get_write_target() const
   return write_lsn + max_buf_free / 2;
 }
 
-template<bool mmap>
+template<bool pmem>
 void mtr_t::commit_log(mtr_t *mtr, std::pair<lsn_t,page_flush_ahead> lsns)
 {
   size_t modified= 0;
-  const lsn_t write_lsn= mmap ? 0 : log_sys.get_write_target();
+  const lsn_t write_lsn= pmem ? 0 : log_sys.get_write_target();
 
   if (mtr->m_made_dirty)
   {
@@ -476,7 +474,7 @@ void mtr_t::commit_log(mtr_t *mtr, std::pair<lsn_t,page_flush_ahead> lsns)
   if (UNIV_UNLIKELY(lsns.second != PAGE_FLUSH_NO))
     buf_flush_ahead(mtr->m_commit_lsn, lsns.second == PAGE_FLUSH_SYNC);
 
-  if (!mmap && UNIV_UNLIKELY(write_lsn != 0))
+  if (!pmem && UNIV_UNLIKELY(write_lsn != 0))
     log_write_up_to(write_lsn, false);
 }
 
@@ -484,7 +482,6 @@ void mtr_t::commit_log(mtr_t *mtr, std::pair<lsn_t,page_flush_ahead> lsns)
 void mtr_t::commit()
 {
   ut_ad(is_active());
-  ut_ad(!is_inside_ibuf());
 
   /* This is a dirty read, for debugging. */
   ut_ad(!m_modifications || !recv_no_log_write);
@@ -562,7 +559,6 @@ inline void fil_space_t::set_create_lsn(lsn_t lsn)
 void mtr_t::commit_shrink(fil_space_t &space, uint32_t size)
 {
   ut_ad(is_active());
-  ut_ad(!is_inside_ibuf());
   ut_ad(!high_level_read_only);
   ut_ad(m_modifications);
   ut_ad(!m_memo.empty());
@@ -668,7 +664,6 @@ void mtr_t::commit_shrink(fil_space_t &space, uint32_t size)
 bool mtr_t::commit_file(fil_space_t &space, const char *name)
 {
   ut_ad(is_active());
-  ut_ad(!is_inside_ibuf());
   ut_ad(!high_level_read_only);
   ut_ad(m_modifications);
   ut_ad(!m_made_dirty);
@@ -748,7 +743,6 @@ ATTRIBUTE_COLD lsn_t mtr_t::commit_files(lsn_t checkpoint_lsn)
 {
   ut_ad(log_sys.latch_have_wr());
   ut_ad(is_active());
-  ut_ad(!is_inside_ibuf());
   ut_ad(m_log_mode == MTR_LOG_ALL);
   ut_ad(!m_made_dirty);
   ut_ad(m_memo.empty());
@@ -897,7 +891,7 @@ ATTRIBUTE_COLD static void log_overwrite_warning(lsn_t lsn)
                   " last checkpoint LSN=" LSN_PF ", current LSN=" LSN_PF
                   "%s.",
                   lsn_t{log_sys.last_checkpoint_lsn}, lsn,
-                  srv_shutdown_state > SRV_SHUTDOWN_INITIATED
+                  srv_shutdown_state != SRV_SHUTDOWN_INITIATED
                   ? ". Shutdown is in progress" : "");
 }
 
@@ -1012,7 +1006,7 @@ ATTRIBUTE_COLD size_t log_t::append_prepare_wait(size_t b, bool ex, lsn_t lsn)
   else
     latch.rd_unlock();
 
-  log_write_up_to(lsn, is_mmap());
+  log_write_up_to(lsn, is_pmem());
 
   if (ex)
     latch.wr_lock(SRW_LOCK_CALL);
@@ -1028,37 +1022,31 @@ ATTRIBUTE_COLD size_t log_t::append_prepare_wait(size_t b, bool ex, lsn_t lsn)
 
 /** Reserve space in the log buffer for appending data.
 @tparam spin  whether to use the spin-only lock_lsn()
-@tparam mmap  log_sys.is_mmap()
+@tparam pmem  log_sys.is_pmem()
 @param size   total length of the data to append(), in bytes
 @param ex     whether log_sys.latch is exclusively locked
 @return the start LSN and the buffer position for append() */
-template<bool spin,bool mmap>
+template<bool spin,bool pmem>
 inline
 std::pair<lsn_t,byte*> log_t::append_prepare(size_t size, bool ex) noexcept
 {
   ut_ad(ex ? latch_have_wr() : latch_have_rd());
-  ut_ad(mmap == is_mmap());
+  ut_ad(pmem == is_pmem());
   if (!spin)
     lsn_lock.wr_lock();
   size_t b{spin ? lock_lsn() : buf_free.load(std::memory_order_relaxed)};
   write_to_buf++;
 
-  lsn_t l{lsn.load(std::memory_order_relaxed)}, end_lsn{l + size};
+  const lsn_t l{lsn.load(std::memory_order_relaxed)}, end_lsn{l + size};
 
-  if (UNIV_UNLIKELY(mmap
+  if (UNIV_UNLIKELY(pmem
                     ? (end_lsn -
                        get_flushed_lsn(std::memory_order_relaxed)) > capacity()
                     : b + size >= buf_size))
-  {
     b= append_prepare_wait<spin>(b, ex, l);
-    /* While flushing log, we had released the lsn lock and LSN could have
-    progressed in the meantime. */
-    l= lsn.load(std::memory_order_relaxed);
-    end_lsn= l + size;
-  }
 
   size_t new_buf_free= b + size;
-  if (mmap && new_buf_free >= file_size)
+  if (pmem && new_buf_free >= file_size)
     new_buf_free-= size_t(capacity());
 
   lsn.store(end_lsn, std::memory_order_relaxed);
@@ -1149,8 +1137,6 @@ std::pair<lsn_t,mtr_t::page_flush_ahead> mtr_t::do_write()
   ut_ad(is_logged());
   ut_ad(m_log.size());
   ut_ad(!m_latch_ex || log_sys.latch_have_wr());
-  ut_ad(!m_user_space ||
-        (m_user_space->id > 0 && m_user_space->id < SRV_SPACE_ID_UPPER_BOUND));
 
 #ifndef DBUG_OFF
   do
@@ -1189,7 +1175,7 @@ std::pair<lsn_t,mtr_t::page_flush_ahead> mtr_t::do_write()
     log_sys.latch.rd_lock(SRW_LOCK_CALL);
 
   if (UNIV_UNLIKELY(m_user_space && !m_user_space->max_lsn &&
-                    !srv_is_undo_tablespace((m_user_space->id))))
+                    !is_predefined_tablespace(m_user_space->id)))
   {
     if (!m_latch_ex)
     {
@@ -1216,47 +1202,23 @@ inline void log_t::resize_write(lsn_t lsn, const byte *end, size_t len,
     end-= len;
     size_t s;
 
-#ifdef HAVE_INNODB_MMAP
+#ifdef HAVE_PMEM
     if (!resize_flush_buf)
     {
-      ut_ad(is_mmap());
-      lsn_lock.wr_lock();
+      ut_ad(is_pmem());
       const size_t resize_capacity{resize_target - START_OFFSET};
+      const lsn_t resizing{resize_in_progress()};
+      if (UNIV_UNLIKELY(lsn < resizing))
       {
-        const lsn_t resizing{resize_in_progress()};
-        /* For memory-mapped log, log_t::resize_start() would never
-        set log_sys.resize_lsn to less than log_sys.lsn. It cannot
-        execute concurrently with this thread, because we are holding
-        log_sys.latch and it would hold an exclusive log_sys.latch. */
-        if (UNIV_UNLIKELY(lsn < resizing))
-        {
-          /* This function may execute in multiple concurrent threads
-          that hold a shared log_sys.latch. Before we got lsn_lock,
-          another thread could have executed resize_lsn.store(lsn) below
-          with a larger lsn than ours.
-
-          append_prepare() guarantees that the concurrent writes
-          cannot overlap, that is, our entire log must be discarded.
-          Besides, incomplete mini-transactions cannot be parsed anyway. */
-          ut_ad(resizing >= lsn + len);
-          goto mmap_done;
-        }
-
-        s= START_OFFSET;
-
-        if (UNIV_UNLIKELY(lsn - resizing + len >= resize_capacity))
-        {
-          resize_lsn.store(lsn, std::memory_order_relaxed);
-          lsn= 0;
-        }
-        else
-        {
-          lsn-= resizing;
-          s+= lsn;
-        }
+        size_t l= resizing - lsn;
+        if (l >= len)
+          return;
+        end+= l - len;
+        len-= l;
+        lsn+= l;
       }
-
-      ut_ad(s + len <= resize_target);
+      lsn-= resizing;
+      s= START_OFFSET + lsn % resize_capacity;
 
       if (UNIV_UNLIKELY(end < &buf[START_OFFSET]))
       {
@@ -1266,22 +1228,59 @@ inline void log_t::resize_write(lsn_t lsn, const byte *end, size_t len,
         ut_ad(end + capacity() + len >= &buf[file_size]);
 
         size_t l= size_t(buf - (end - START_OFFSET));
-        memcpy(resize_buf + s, end + capacity(), l);
-        memcpy(resize_buf + s + l, &buf[START_OFFSET], len - l);
+        if (UNIV_LIKELY(s + len <= resize_target))
+        {
+          /* The destination buffer (log_sys.resize_buf) did not wrap around */
+          memcpy(resize_buf + s, end + capacity(), l);
+          memcpy(resize_buf + s + l, &buf[START_OFFSET], len - l);
+          goto pmem_nowrap;
+        }
+        else
+        {
+          /* Both log_sys.buf and log_sys.resize_buf wrapped around */
+          const size_t rl= resize_target - s;
+          if (l <= rl)
+          {
+            /* log_sys.buf wraps around first */
+            memcpy(resize_buf + s, end + capacity(), l);
+            memcpy(resize_buf + s + l, &buf[START_OFFSET], rl - l);
+            memcpy(resize_buf + START_OFFSET, &buf[START_OFFSET + rl - l],
+                   len - l);
+          }
+          else
+          {
+            /* log_sys.resize_buf wraps around first */
+            memcpy(resize_buf + s, end + capacity(), rl);
+            memcpy(resize_buf + START_OFFSET, end + capacity() + rl, l - rl);
+            memcpy(resize_buf + START_OFFSET + (l - rl),
+                   &buf[START_OFFSET], len - l);
+          }
+          goto pmem_wrap;
+        }
       }
       else
       {
         ut_ad(end + len <= &buf[file_size]);
-        memcpy(resize_buf + s, end, len);
-      }
-      s+= len - seq;
 
-      /* Always set the sequence bit. If the resized log were to wrap around,
-      we will advance resize_lsn. */
-      ut_ad(resize_buf[s] <= 1);
-      resize_buf[s]= 1;
-    mmap_done:
-      lsn_lock.wr_unlock();
+        if (UNIV_LIKELY(s + len <= resize_target))
+        {
+          memcpy(resize_buf + s, end, len);
+        pmem_nowrap:
+          s+= len - seq;
+        }
+        else
+        {
+          /* The log_sys.resize_buf wrapped around */
+          memcpy(resize_buf + s, end, resize_target - s);
+          memcpy(resize_buf + START_OFFSET, end + (resize_target - s),
+                 len - (resize_target - s));
+        pmem_wrap:
+          s+= len - seq;
+          if (s >= resize_target)
+            s-= resize_capacity;
+          resize_lsn.fetch_add(resize_capacity); /* Move the target ahead. */
+        }
+      }
     }
     else
 #endif
@@ -1291,24 +1290,16 @@ inline void log_t::resize_write(lsn_t lsn, const byte *end, size_t len,
       ut_ad(s + len <= buf_size);
       memcpy(resize_buf + s, end, len);
       s+= len - seq;
-      /* Always set the sequence bit. If the resized log were to wrap around,
-      we will advance resize_lsn. */
-      ut_ad(resize_buf[s] <= 1);
-      resize_buf[s]= 1;
     }
+
+    /* Always set the sequence bit. If the resized log were to wrap around,
+    we will advance resize_lsn. */
+    ut_ad(resize_buf[s] <= 1);
+    resize_buf[s]= 1;
   }
 }
 
-inline void log_t::append(byte *&d, const void *s, size_t size) noexcept
-{
-  ut_ad(log_sys.latch_have_any());
-  ut_ad(d + size <= log_sys.buf +
-        (log_sys.is_mmap() ? log_sys.file_size : log_sys.buf_size));
-  memcpy(d, s, size);
-  d+= size;
-}
-
-template<bool spin,bool mmap>
+template<bool spin,bool pmem>
 std::pair<lsn_t,mtr_t::page_flush_ahead>
 mtr_t::finish_writer(mtr_t *mtr, size_t len)
 {
@@ -1319,14 +1310,16 @@ mtr_t::finish_writer(mtr_t *mtr, size_t len)
 
   const size_t size{mtr->m_commit_lsn ? 5U + 8U : 5U};
   std::pair<lsn_t, byte*> start=
-    log_sys.append_prepare<spin,mmap>(len, mtr->m_latch_ex);
+    log_sys.append_prepare<spin,pmem>(len, mtr->m_latch_ex);
 
-  if (!mmap)
+  if (!pmem)
   {
     mtr->m_log.for_each_block([&start](const mtr_buf_t::block_t *b)
     { log_sys.append(start.second, b->begin(), b->used()); return true; });
 
+#ifdef HAVE_PMEM
   write_trailer:
+#endif
     *start.second++= log_sys.get_sequence_bit(start.first + len - size);
     if (mtr->m_commit_lsn)
     {
@@ -1337,6 +1330,7 @@ mtr_t::finish_writer(mtr_t *mtr, size_t len)
     mach_write_to_4(start.second, mtr->m_crc);
     start.second+= 4;
   }
+#ifdef HAVE_PMEM
   else
   {
     if (UNIV_LIKELY(start.second + len <= &log_sys.buf[log_sys.file_size]))
@@ -1384,6 +1378,9 @@ mtr_t::finish_writer(mtr_t *mtr, size_t len)
       ((size >= size_left) ? log_sys.START_OFFSET : log_sys.file_size) +
       (size - size_left);
   }
+#else
+  static_assert(!pmem, "");
+#endif
 
   log_sys.resize_write(start.first, start.second, len, size);
 
