@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1996, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2015, 2022, MariaDB Corporation.
+Copyright (c) 2015, 2023, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -114,8 +114,6 @@ trx_init(
 
 	trx->lock.n_rec_locks = 0;
 
-	trx->lock.set_nth_bit_calls = 0;
-
 	trx->dict_operation = false;
 
 	trx->error_state = DB_SUCCESS;
@@ -355,7 +353,6 @@ trx_t *trx_create()
 
 	ut_ad(trx->mod_tables.empty());
 	ut_ad(trx->lock.n_rec_locks == 0);
-	ut_ad(trx->lock.set_nth_bit_calls == 0);
 	ut_ad(trx->lock.table_cached == 0);
 	ut_ad(trx->lock.rec_cached == 0);
 	ut_ad(UT_LIST_GET_LEN(trx->lock.evicted_tables) == 0);
@@ -541,7 +538,6 @@ TRANSACTIONAL_TARGET void trx_free_at_shutdown(trx_t *trx)
 	DBUG_LOG("trx", "Free prepared: " << trx);
 	trx->state = TRX_STATE_NOT_STARTED;
 	ut_ad(!UT_LIST_GET_LEN(trx->lock.trx_locks));
-	ut_d(*trx->detailed_error = '\0');
 	trx->free();
 }
 
@@ -926,7 +922,6 @@ trx_start_low(
 
 #ifdef WITH_WSREP
 	trx->xid.null();
-	trx->wsrep = wsrep_on(trx->mysql_thd);
 #endif /* WITH_WSREP */
 
 	ut_a(ib_vector_is_empty(trx->autoinc_locks));
@@ -1261,22 +1256,22 @@ static void trx_flush_log_if_needed(lsn_t lsn, trx_t *trx)
     return;
 
   const bool flush=
-    (srv_file_flush_method != SRV_NOSYNC &&
+    (!my_disable_sync &&
      (srv_flush_log_at_trx_commit & 1));
-  if (!log_sys.is_mmap())
-  {
-    completion_callback cb;
 
-    if ((cb.m_param= thd_increment_pending_ops(trx->mysql_thd)))
-    {
-      cb.m_callback= (void (*)(void *)) thd_decrement_pending_ops;
-      log_write_up_to(lsn, flush, &cb);
-      return;
-    }
+  completion_callback cb;
+  if (!log_sys.is_pmem() &&
+      (cb.m_param= thd_increment_pending_ops(trx->mysql_thd)))
+  {
+    cb.m_callback = (void (*)(void *)) thd_decrement_pending_ops;
+    log_write_up_to(lsn, flush, &cb);
+  }
+  else
+  {
+    trx->op_info= "flushing log";
+    log_write_up_to(lsn, flush);
+    trx->op_info= "";
   }
-  trx->op_info= "flushing log";
-  log_write_up_to(lsn, flush);
-  trx->op_info= "";
 }
 
 /** Process tables that were modified by the committing transaction. */
@@ -1501,8 +1496,6 @@ TRANSACTIONAL_INLINE inline void trx_t::commit_in_memory(const mtr_t *mtr)
     trx_finalize_for_fts(this, undo_no != 0);
 
 #ifdef WITH_WSREP
-  ut_ad(is_wsrep() == wsrep_on(mysql_thd));
-
   /* Serialization history has been written and the transaction is
   committed in memory, which makes this commit ordered. Release commit
   order critical section. */
@@ -1825,7 +1818,7 @@ trx_print_low(
 			/*!< in: mem_heap_get_size(trx->lock.lock_heap) */
 {
 	if (const trx_id_t id = trx->id) {
-		fprintf(f, "TRANSACTION " TRX_ID_FMT, id);
+		fprintf(f, "TRANSACTION " TRX_ID_FMT, trx->id);
 	} else {
 		fprintf(f, "TRANSACTION (%p)", trx);
 	}
@@ -2026,8 +2019,7 @@ trx_prepare(
 
 		We must not be holding any mutexes or latches here. */
 		if (auto f = srv_flush_log_at_trx_commit) {
-			log_write_up_to(lsn, (f & 1) && srv_file_flush_method
-					!= SRV_NOSYNC);
+			log_write_up_to(lsn, (f & 1) && !my_disable_sync);
 		}
 
 		if (!UT_LIST_GET_LEN(trx->lock.trx_locks)
@@ -2104,9 +2096,9 @@ static my_bool trx_recover_for_mysql_callback(rw_trx_hash_element_t *element,
 }
 
 
-static my_bool trx_recover_reset_callback(void *el, void*)
+static my_bool trx_recover_reset_callback(rw_trx_hash_element_t *element,
+  void*)
 {
-  rw_trx_hash_element_t *element= static_cast<rw_trx_hash_element_t*>(el);
   element->mutex.wr_lock();
   if (trx_t *trx= element->trx)
   {
@@ -2158,10 +2150,9 @@ struct trx_get_trx_by_xid_callback_arg
 };
 
 
-static my_bool trx_get_trx_by_xid_callback(void *el, void *a)
+static my_bool trx_get_trx_by_xid_callback(rw_trx_hash_element_t *element,
+  trx_get_trx_by_xid_callback_arg *arg)
 {
-  auto element= static_cast<rw_trx_hash_element_t*>(el);
-  auto arg= static_cast<trx_get_trx_by_xid_callback_arg*>(a);
   my_bool found= 0;
   element->mutex.wr_lock();
   if (trx_t *trx= element->trx)
